"""
Class for reading data from from Tucker Davis TTank format.
Terminology:
TDT hold data with tanks (actually a directory). And tanks hold sub block
(sub directories).
Tanks correspond to neo.Block and tdt block correspond to neo.Segment.

Note the name Block is ambiguous because it does not refer to same thing in TDT
terminology and neo.


In a directory there are several files:
  * TSQ timestamp index of data
  * TBK some kind of channel info and maybe more
  * TEV contains data : spike + event + signal (for old version)
  * SEV contains signals (for new version)
  * ./sort/ can contain offline spikesorting label for spike
     and can be use place of TEV.

Units in this IO are not guaranteed.

Author: Samuel Garcia, SummitKwan, Chadwick Boulay

Alternative package for loading the tdt format:
https://pypi.org/project/tdt
"""

import numpy as np
import os
import re
import warnings
from collections import OrderedDict
from pathlib import Path

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

from neo.core import NeoReadWriteError


class TdtRawIO(BaseRawIO):
    extensions = ["tbk", "tdx", "tev", "tin", "tnt", "tsq", "sev", "txt"]
    rawmode = "one-dir"

    def __init__(self, dirname="", sortname=""):
        """
        Class for reading  one or multiple TDT data blocks.

        Parameters
        ----------
        dirname: str | Path, default: None
            Tank-directory of a dataset to be read as multiple segments OR single file of dataset.
            In the latter case only the corresponding segment will be considered.
        sortname: str, default: ''
            'sortname' is used to specify the external sortcode generated by offline spike sorting.
            if sortname=='PLX', there should be a ./sort/PLX/*.SortResult file in the tdt block,
            which stores the sortcode for every spike
            '' will use original online sorting

        Notes
        -----
        In a directory there are several files:
         * TSQ timestamp index of data
         * TBK some kind of channel info and maybe more
         * TEV contains data : spike + event + signal (for old version)
         * SEV contains signals (for new version)
         * ./sort/ can contain offline spikesorting label for spike and can be use place of TEV.

        """
        BaseRawIO.__init__(self)
        dirname = Path(dirname)
        if dirname.is_dir():
            self.dirname = Path(dirname)
            self.tdt_block_mode = "multi"
        elif dirname.is_file():
            # in single tdt block mode the dirname also contains the block prefix
            self.dirname = dirname.with_suffix("")
            self.tdt_block_mode = "single"
        else:
            raise ValueError(f"No data folder or file found for {dirname}")

        self.sortname = sortname

    def _source_name(self):
        return self.dirname

    def _get_filestem(self, segment_name=""):
        if self.tdt_block_mode == "multi":
            return self.dirname / segment_name / f"{self.dirname.name}_{segment_name}"
        else:
            return self.dirname

    def _parse_header(self):
        segment_names = []
        if self.tdt_block_mode == "multi":
            tankname = self.dirname.stem
            for path in self.dirname.iterdir():
                if is_tdtblock(path):
                    segment_names.append(path.stem)

        # if no block structure was detected, check if current dir contains a set of data
        elif is_tdtblock(self.dirname.parent):
            segment_names.append(str(self.dirname.stem))
            tankname = None
        else:
            raise ValueError(
                f"{self.dirname} is not a valid tdt structure. Make sure all files "
                f"required for a tdt block are present."
            )

        nb_segment = len(segment_names)
        if nb_segment == 0:
            warnings.warn(f"Could not find any data set belonging to {self.dirname}")

        # TBK (channel info)
        info_channel_groups = None
        for seg_index, segment_name in enumerate(segment_names):

            # TBK contain channels
            tbk_filename = self._get_filestem(segment_name).with_suffix(".Tbk")
            _info_channel_groups = read_tbk(tbk_filename)
            if info_channel_groups is None:
                info_channel_groups = _info_channel_groups
            else:
                if not np.array_equal(info_channel_groups, _info_channel_groups):
                    raise NeoReadWriteError("Channels differ across segments")

        # TEV (mixed data)
        self._tev_datas = []
        for seg_index, segment_name in enumerate(segment_names):
            tev_filename = self._get_filestem(segment_name).with_suffix(".tev")
            if tev_filename.exists():
                tev_data = np.memmap(tev_filename, mode="r", offset=0, dtype="uint8")
            else:
                tev_data = None
            self._tev_datas.append(tev_data)

        # TSQ index with timestamp
        self._tsq = []
        self._seg_t_starts = []
        self._seg_t_stops = []
        for seg_index, segment_name in enumerate(segment_names):
            tsq_filename = self._get_filestem(segment_name).with_suffix(".tsq")
            tsq = np.fromfile(tsq_filename, dtype=tsq_dtype)
            self._tsq.append(tsq)
            # Start and stop times are only found in the second
            #  and last header row, respectively.
            if tsq[1]["evname"] == chr(EVMARK_STARTBLOCK).encode():
                self._seg_t_starts.append(tsq[1]["timestamp"])
            else:
                self._seg_t_starts.append(np.nan)
                print("segment start time not found")
            if tsq[-1]["evname"] == chr(EVMARK_STOPBLOCK).encode():
                self._seg_t_stops.append(tsq[-1]["timestamp"])
            else:
                self._seg_t_stops.append(np.nan)
                print("segment stop time not found")

            # If there exists an external sortcode in ./sort/[sortname]/*.SortResult
            #  (generated after offline sorting)
            if self.sortname != "":
                try:
                    if self.tdt_block_mode == "multi":
                        path = self.dirname
                    else:
                        path = self.dirname.parent
                    for file in os.listdir(path / "sort" / self.sortname):
                        if file.endswith(".SortResult"):
                            sortresult_filename = path / "sort" / self.sortname / file
                            # get new sortcode
                            newsortcode = np.fromfile(sortresult_filename, "int8")[1024:]  # first 1024 bytes are header
                            # update the sort code with the info from this file
                            tsq["sortcode"][1:-1] = newsortcode
                            # print('sortcode updated')
                            break
                except OSError:
                    pass

        # Re-order segments according to their start times
        sort_inds = np.argsort(self._seg_t_starts)
        if not np.array_equal(sort_inds, list(range(nb_segment))):
            segment_names = [segment_names[x] for x in sort_inds]
            self._tev_datas = [self._tev_datas[x] for x in sort_inds]
            self._seg_t_starts = [self._seg_t_starts[x] for x in sort_inds]
            self._seg_t_stops = [self._seg_t_stops[x] for x in sort_inds]
            self._tsq = [self._tsq[x] for x in sort_inds]
        self._global_t_start = self._seg_t_starts[0]

        # signal channels EVTYPE_STREAM
        signal_streams = []
        signal_channels = []
        self._sigs_data_buf = {seg_index: {} for seg_index in range(nb_segment)}
        self._sigs_index = {seg_index: {} for seg_index in range(nb_segment)}
        self._sig_dtype_by_group = {}  # key = group_id
        self._sig_sample_per_chunk = {}  # key = group_id
        self._sigs_lengths = {seg_index: {} for seg_index in range(nb_segment)}  # key = seg_index then group_id
        self._sigs_t_start = {seg_index: {} for seg_index in range(nb_segment)}  # key = seg_index then group_id

        keep = info_channel_groups["TankEvType"] & EVTYPE_MASK == EVTYPE_STREAM
        missing_sev_channels = []
        for stream_index, info in enumerate(info_channel_groups[keep]):
            self._sig_sample_per_chunk[stream_index] = info["NumPoints"]

            stream_name = str(info["StoreName"])
            stream_id = f"{stream_index}"
            signal_streams.append((stream_name, stream_id))

            for c in range(info["NumChan"]):
                global_chan_index = len(signal_channels)
                chan_id = c + 1  # several StoreName can have same chan_id: this is ok

                # loop over segment to get sampling_rate/data_index/data_buffer
                sampling_rate = None
                dtype = None
                for seg_index, segment_name in enumerate(segment_names):
                    # get data index
                    tsq = self._tsq[seg_index]
                    mask = (
                        (tsq["evtype"] & EVTYPE_MASK == EVTYPE_STREAM)
                        & (tsq["evname"] == info["StoreName"])
                        & (tsq["channel"] == chan_id)
                    )
                    data_index = tsq[mask].copy()
                    self._sigs_index[seg_index][global_chan_index] = data_index

                    size = info["NumPoints"] * data_index.size
                    if stream_index not in self._sigs_lengths[seg_index]:
                        self._sigs_lengths[seg_index][stream_index] = size
                    else:
                        if self._sigs_lengths[seg_index][stream_index] != size:
                            raise ValueError(f"The sig lengths for {seg_index=}, {stream_index=} is not of size {size}")

                    # signal start time, relative to start of segment
                    if len(data_index["timestamp"]):
                        t_start = data_index["timestamp"][0]
                    else:
                        # if no signal present use segment t_start as dummy value
                        t_start = self._seg_t_starts[seg_index]
                    if stream_index not in self._sigs_t_start[seg_index]:
                        self._sigs_t_start[seg_index][stream_index] = t_start
                    else:
                        if self._sigs_t_start[seg_index][stream_index] != t_start:
                            raise ValueError(f"The t_start for {seg_index=}, {stream_index=} is not of time {t_start}")

                    # sampling_rate and dtype
                    if len(data_index):
                        _sampling_rate = float(data_index["frequency"][0])
                        _dtype = data_formats[data_index["dataformat"][0]]
                    else:
                        # if no signal present use dummy values
                        _sampling_rate = 1.0
                        _dtype = int
                    if sampling_rate is None:
                        sampling_rate = _sampling_rate
                        dtype = _dtype
                        if stream_index not in self._sig_dtype_by_group:
                            self._sig_dtype_by_group[stream_index] = np.dtype(dtype)
                        else:
                            if self._sig_dtype_by_group[stream_index] != dtype:
                                raise TypeError(
                                    f"The dtype for the signal in {stream_index=} is not the correct dtype of {dtype}"
                                )
                    else:
                        if sampling_rate != _sampling_rate:
                            raise ValueError("sampling is changing!!!")
                        if dtype != _dtype:
                            raise ValueError("Dtype is changing!!")

                    # data buffer test if SEV file exists otherwise TEV
                    # path = self.dirname / segment_name
                    if self.tdt_block_mode == "multi":
                        # for multi block datasets the names of sev files are fixed
                        store = info["StoreName"].decode("ascii")
                        sev_stem = f"{tankname}_{segment_name}_{store}_ch{chan_id}"
                        sev_filename = (path / sev_stem).with_suffix(".sev")
                    else:
                        # for single block datasets the exact name of sev files in not known
                        sev_regex = f"*_[cC]h{chan_id}.sev"
                        sev_filename = list(self.dirname.parent.glob(str(sev_regex)))
                        # in case multiple sev files are found, try to find the one for current stream
                        if len(sev_filename) > 1:
                            store = info["StoreName"].decode("ascii")
                            sev_regex = f"*_{store}_Ch{chan_id}.sev"
                            sev_filename = list(self.dirname.parent.glob(str(sev_regex)))

                        # in case non or multiple sev files are found for current stream + channel
                        if len(sev_filename) != 1:
                            missing_sev_channels.append(chan_id)
                            sev_filename = None
                        else:
                            sev_filename = sev_filename[0]

                    if (sev_filename is not None) and sev_filename.exists():
                        data = np.memmap(sev_filename, mode="r", offset=0, dtype="uint8")
                    else:
                        data = self._tev_datas[seg_index]
                    if data is None:
                        raise NeoReadWriteError("no TEV nor SEV data to read")
                    self._sigs_data_buf[seg_index][global_chan_index] = data

                chan_name = f"{info['StoreName']} {c + 1}"
                sampling_rate = sampling_rate
                units = "uV"  # see https://github.com/NeuralEnsemble/python-neo/issues/1369
                gain = 1.0
                offset = 0.0
                signal_channels.append((chan_name, str(chan_id), sampling_rate, dtype, units, gain, offset, stream_id))

        if missing_sev_channels:
            warnings.warn(f"Could not identify sev files for channels {missing_sev_channels}.")

        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # unit channels EVTYPE_SNIP
        self.internal_unit_ids = {}
        self._waveforms_size = []
        self._waveforms_dtype = []
        spike_channels = []
        keep = info_channel_groups["TankEvType"] == EVTYPE_SNIP
        tsq = np.hstack(self._tsq)
        # If there is no chance the different TSQ files will have different units,
        #  then we can do tsq = self._tsq[0]
        for info in info_channel_groups[keep]:
            for c in range(info["NumChan"]):
                chan_id = c + 1
                mask = (
                    (tsq["evtype"] == EVTYPE_SNIP) & (tsq["evname"] == info["StoreName"]) & (tsq["channel"] == chan_id)
                )
                unit_ids = np.unique(tsq[mask]["sortcode"])
                for unit_id in unit_ids:
                    unit_index = len(spike_channels)
                    self.internal_unit_ids[unit_index] = (info["StoreName"], chan_id, unit_id)

                    unit_name = f"ch{chan_id}#{unit_id}"
                    wf_units = "V"
                    wf_gain = 1.0
                    wf_offset = 0.0
                    wf_left_sweep = info["NumPoints"] // 2
                    wf_sampling_rate = info["SampleFreq"]
                    spike_channels.append(
                        (unit_name, f"{unit_id}", wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate)
                    )

                    self._waveforms_size.append(info["NumPoints"])
                    self._waveforms_dtype.append(np.dtype(data_formats[info["DataFormat"]]))

        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # signal channels EVTYPE_STRON
        event_channels = []
        keep = info_channel_groups["TankEvType"] == EVTYPE_STRON
        for info in info_channel_groups[keep]:
            chan_name = info["StoreName"]
            chan_id = 1
            event_channels.append((chan_name, chan_id, "event"))

        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # fill into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Annotations only standard ones:
        self._generate_minimal_annotations()

    def _block_count(self):
        return 1

    def _segment_count(self, block_index):
        return self.header["nb_segment"][block_index]

    def _segment_t_start(self, block_index, seg_index):
        return self._seg_t_starts[seg_index] - self._global_t_start

    def _segment_t_stop(self, block_index, seg_index):
        return self._seg_t_stops[seg_index] - self._global_t_start

    def _get_signal_size(self, block_index, seg_index, stream_index):
        size = self._sigs_lengths[seg_index][stream_index]
        return size

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return self._sigs_t_start[seg_index][stream_index] - self._global_t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._sigs_lengths[seg_index][stream_index]

        stream_id = self.header["signal_streams"][stream_index]["id"]
        signal_channels = self.header["signal_channels"]
        mask = signal_channels["stream_id"] == stream_id
        global_chan_indexes = np.arange(signal_channels.size)[mask]
        signal_channels = signal_channels[mask]

        if channel_indexes is None:
            channel_indexes = slice(None)
        global_chan_indexes = global_chan_indexes[channel_indexes]
        signal_channels = signal_channels[channel_indexes]

        dt = self._sig_dtype_by_group[stream_index]
        raw_signals = np.zeros((i_stop - i_start, signal_channels.size), dtype=dt)

        sample_per_chunk = self._sig_sample_per_chunk[stream_index]
        bl0 = i_start // sample_per_chunk
        bl1 = int(np.ceil(i_stop / sample_per_chunk))
        chunk_nb_bytes = sample_per_chunk * dt.itemsize

        for c, global_index in enumerate(global_chan_indexes):
            data_index = self._sigs_index[seg_index][global_index]
            data_buf = self._sigs_data_buf[seg_index][global_index]

            # loop over data blocks and get chunks
            ind = 0
            for bl in range(bl0, bl1):
                ind0 = data_index[bl]["offset"]
                ind1 = ind0 + chunk_nb_bytes
                data = data_buf[ind0:ind1].view(dt)

                if bl == bl1 - 1:
                    # right border
                    # be careful that bl could be both bl0 and bl1!!
                    border = data.size - (i_stop % sample_per_chunk)
                    # cut data if not all samples are requested
                    if border != len(data):
                        data = data[:-border]
                if bl == bl0:
                    # left border
                    border = i_start % sample_per_chunk
                    data = data[border:]

                raw_signals[ind : data.size + ind, c] = data
                ind += data.size

        return raw_signals

    def _get_mask(self, tsq, seg_index, evtype, evname, chan_id, unit_id, t_start, t_stop):
        """Used inside spike and events methods"""
        mask = (tsq["evtype"] == evtype) & (tsq["evname"] == evname) & (tsq["channel"] == chan_id)

        if unit_id is not None:
            mask &= tsq["sortcode"] == unit_id

        if t_start is not None:
            mask &= tsq["timestamp"] >= (t_start + self._global_t_start)

        if t_stop is not None:
            mask &= tsq["timestamp"] <= (t_stop + self._global_t_start)

        return mask

    def _spike_count(self, block_index, seg_index, unit_index):
        store_name, chan_id, unit_id = self.internal_unit_ids[unit_index]
        tsq = self._tsq[seg_index]
        mask = self._get_mask(tsq, seg_index, EVTYPE_SNIP, store_name, chan_id, unit_id, None, None)
        nb_spike = np.sum(mask)
        return nb_spike

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        store_name, chan_id, unit_id = self.internal_unit_ids[unit_index]
        tsq = self._tsq[seg_index]
        mask = self._get_mask(tsq, seg_index, EVTYPE_SNIP, store_name, chan_id, unit_id, t_start, t_stop)
        timestamps = tsq[mask]["timestamp"]
        timestamps -= self._global_t_start
        return timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        # already in s
        spike_times = spike_timestamps.astype(dtype)
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        store_name, chan_id, unit_id = self.internal_unit_ids[unit_index]
        tsq = self._tsq[seg_index]
        mask = self._get_mask(tsq, seg_index, EVTYPE_SNIP, store_name, chan_id, unit_id, t_start, t_stop)
        nb_spike = np.sum(mask)

        data = self._tev_datas[seg_index]

        dt = self._waveforms_dtype[unit_index]
        nb_sample = self._waveforms_size[unit_index]
        waveforms = np.zeros((nb_spike, 1, nb_sample), dtype=dt)

        for i, e in enumerate(tsq[mask]):
            ind0 = e["offset"]
            ind1 = ind0 + nb_sample * dt.itemsize
            waveforms[i, 0, :] = data[ind0:ind1].view(dt)

        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        h = self.header["event_channels"][event_channel_index]
        store_name = h["name"].encode("ascii")
        tsq = self._tsq[seg_index]
        chan_id = 0
        mask = self._get_mask(tsq, seg_index, EVTYPE_STRON, store_name, chan_id, None, None, None)
        nb_event = np.sum(mask)
        return nb_event

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        h = self.header["event_channels"][event_channel_index]
        store_name = h["name"].encode("ascii")
        tsq = self._tsq[seg_index]
        chan_id = 0
        mask = self._get_mask(tsq, seg_index, EVTYPE_STRON, store_name, chan_id, None, None, None)

        timestamps = tsq[mask]["timestamp"]
        timestamps -= self._global_t_start
        labels = tsq[mask]["offset"].astype("U")
        durations = None
        # TODO if user demand event to epoch
        # with EVTYPE_STROFF=258
        # and so durations would be not None
        # it was not implemented in previous IO.
        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        # already in s
        ev_times = event_timestamps.astype(dtype)
        return ev_times


tbk_field_types = [
    ("StoreName", "S4"),
    ("HeadName", "S16"),
    ("Enabled", "bool"),
    ("CircType", "int"),
    ("NumChan", "int"),
    ("StrobeMode", "int"),
    ("TankEvType", "int32"),
    ("NumPoints", "int"),
    ("DataFormat", "int"),
    ("SampleFreq", "float64"),
]


def read_tbk(tbk_filename):
    """
    Tbk contains some visible header in txt mode to describe
    channel group info.
    """
    with open(tbk_filename, mode="rb") as f:
        txt_header = f.read()

    infos = []
    for chan_grp_header in txt_header.split(b"[STOREHDRITEM]"):
        if chan_grp_header.startswith(b"[USERNOTEDELIMITER]"):
            break

        # parse into a dict
        info = OrderedDict()
        # name and type have to be words, but value can contain floating numbers
        pattern = rb"NAME=(\w+);TYPE=(\w+);VALUE=(\S+);"
        r = re.findall(pattern, chan_grp_header)
        for name, _type, value in r:
            info[name.decode("ascii")] = value
        infos.append(info)

    # and put into numpy
    tbk_field_names = [tbk_field_type[0] for tbk_field_type in tbk_field_types]
    info_channel_groups = np.zeros(len(infos), dtype=tbk_field_types)
    for i, info in enumerate(infos):
        missing_keys = set(tbk_field_names) - set(info.keys())
        if missing_keys:
            warnings.warn(
                f"The tbk file contains incomplete channel group info for group {list(info.items())}. "
                "This channel group will be skipped."
            )
            continue
        for k, dt in tbk_field_types:
            v = np.dtype(dt).type(info[k])
            info_channel_groups[i][k] = v

    return info_channel_groups


tsq_dtype = [
    ("size", "int32"),  # bytes 0-4
    ("evtype", "int32"),  # bytes 5-8
    ("evname", "S4"),  # bytes 9-12
    ("channel", "uint16"),  # bytes 13-14
    ("sortcode", "uint16"),  # bytes 15-16
    ("timestamp", "float64"),  # bytes 17-24
    ("offset", "int64"),  # bytes 25-32
    ("dataformat", "int32"),  # bytes 33-36
    ("frequency", "float32"),  # bytes 37-40
]

EVTYPE_UNKNOWN = int("00000000", 16)  # 0
EVTYPE_STRON = int("00000101", 16)  # 257
EVTYPE_STROFF = int("00000102", 16)  # 258
EVTYPE_SCALAR = int("00000201", 16)  # 513
EVTYPE_STREAM = int("00008101", 16)  # 33025
EVTYPE_SNIP = int("00008201", 16)  # 33281
EVTYPE_MARK = int("00008801", 16)  # 34817
EVTYPE_HASDATA = int("00008000", 16)  # 32768
EVTYPE_UCF = int("00000010", 16)  # 16
EVTYPE_PHANTOM = int("00000020", 16)  # 32
EVTYPE_MASK = int("0000FF0F", 16)  # 65295
EVTYPE_INVALID_MASK = int("FFFF0000", 16)  # 4294901760
EVMARK_STARTBLOCK = int("0001", 16)  # 1
EVMARK_STOPBLOCK = int("0002", 16)  # 2

data_formats = {
    0: "float32",
    1: "int32",
    2: "int16",
    3: "int8",
    4: "float64",
}


def is_tdtblock(blockpath):
    """Is the path a TDT block (=neo.Segment) ?"""
    file_ext = list()
    if blockpath.is_dir():
        # for every file, get extension, convert to lowercase and append
        for file in blockpath.iterdir():
            file_ext.append(file.suffix.lower())

    file_ext = set(file_ext)
    tdt_ext = {".tbk", ".tdx", ".tev", ".tsq"}
    if file_ext >= tdt_ext:  # if containing all the necessary files
        return True
    else:
        return False



import unittest
from pathlib import Path
from numpy.testing import assert_array_equal, assert_

from neo.rawio.tdtrawio import TdtRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestTdtRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = TdtRawIO
    entities_to_download = ["tdt"]
    entities_to_test = [
        # test structure directory with multiple blocks
        "tdt/aep_05",
        # test single block
        "tdt/dataset_0_single_block/512ch_reconly_all-181123_B24_rest.Tdx",
        "tdt/dataset_1_single_block/ECTest-220207-135355_ECTest_B1.Tdx",
        "tdt/aep_05/Block-1/aep_05_Block-1.Tdx",
    ]

    def test_invalid_dirname(self):
        invalid_name = "random_non_existant_tdt_filename"
        assert not Path(invalid_name).exists()

        with self.assertRaises(ValueError):
            TdtRawIO(invalid_name)

    def test_compare_load_multi_single_block(self):
        dirname = self.get_local_path("tdt/aep_05")
        filename = self.get_local_path("tdt/aep_05/Block-1/aep_05_Block-1.Tdx")

        io_single = TdtRawIO(filename)
        io_multi = TdtRawIO(dirname)

        io_single.parse_header()
        io_multi.parse_header()

        self.assertEqual(io_single.tdt_block_mode, "single")
        self.assertEqual(io_multi.tdt_block_mode, "multi")

        self.assertEqual(io_single.block_count(), 1)
        self.assertEqual(io_multi.block_count(), 1)

        self.assertEqual(io_single.segment_count(0), 1)
        self.assertEqual(io_multi.segment_count(0), 2)

        # compare header infos
        assert_array_equal(io_single.header["signal_streams"], io_multi.header["signal_streams"])
        assert_array_equal(io_single.header["signal_channels"], io_multi.header["signal_channels"])
        assert_array_equal(io_single.header["event_channels"], io_multi.header["event_channels"])

        # not all spiking channels are present in first tdt block (segment)
        for spike_channel in io_single.header["spike_channels"]:
            self.assertIn(spike_channel, io_multi.header["spike_channels"])

        # check that extracted signal chunks are identical
        assert_array_equal(
            io_single.get_analogsignal_chunk(0, 0, 0, 100, 0), io_multi.get_analogsignal_chunk(0, 0, 0, 100, 0)
        )


if __name__ == "__main__":
    unittest.main()



"""
Converting to NWB using Neo
===========================

"""

from urllib.request import urlretrieve
from urllib.parse import quote
from neo.io import get_io


dataset_url = "https://object.cscs.ch/v1/AUTH_63ea6845b1d34ad7a43c8158d9572867/hbp-d000017_PatchClamp-GranuleCells_pub"

folder = "GrC_Subject15_180116"

# filenames = ["180116_0004 IV -70.abf", "180116_0005 CC step.abf", "180116_0006 EPSP.abf"]
filenames = ["180116_0005 CC step.abf"]

for filename in filenames:
    datafile_url = f"{dataset_url}/{folder}/{quote(filename)}"
    local_file = urlretrieve(datafile_url, filename)


reader = get_io("180116_0005 CC step.abf")
data = reader.read()

global_metadata = {
    "session_start_time": data[0].rec_datetime,
    "identifier": data[0].file_origin,
    "session_id": "180116_0005",
    "institution": "University of Pavia",
    "lab": "D'Angelo Lab",
    "related_publications": "https://doi.org/10.1038/s42003-020-0953-x",
}

# data[0].annotate(**global_metadata)

signal_metadata = {
    "nwb_group": "acquisition",
    "nwb_neurodata_type": ("pynwb.icephys", "PatchClampSeries"),
    "nwb_electrode": {
        "name": "patch clamp electrode",
        "description": "The patch-clamp pipettes were pulled from borosilicate glass capillaries "
        "(Hilgenberg, Malsfeld, Germany) and filled with intracellular solution "
        "(K-gluconate based solution)",
        "device": {"name": "patch clamp electrode"},
    },
    "nwb:gain": 1.0,
}

for segment in data[0].segments:
    signal = segment.analogsignals[0]
    signal.annotate(**signal_metadata)

from neo.io import NWBIO

writer = NWBIO("GrC_Subject15_180116.nwb", mode="w", **global_metadata)
writer.write(data)



"""
RawIO for reading EDF and EDF+ files using pyedflib

PyEDFLib
https://pyedflib.readthedocs.io
https://github.com/holgern/pyedflib

EDF Format Specifications: https://www.edfplus.info/

Author: Julia Sprenger
"""

import numpy as np

try:
    from pyedflib import EdfReader

    HAS_PYEDF = True
except ImportError:
    HAS_PYEDF = False

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class EDFRawIO(BaseRawIO):
    """
    Class for reading European Data Format files (EDF and EDF+).
    Currently only continuous EDF+ files (EDF+C) and original EDF files (EDF) are supported

    Parameters
    ----------
    filename: str, default: ''
        The EDF+C file or EDF file to be loaded

    Examples
    --------
    >>> import neo.rawio
    >>> r = neo.rawio.EDFRawIO(filename='file.edf')
    >>> r.parse_header()
    >>> print(r)
    >>> raw_chunk = r.get_analogsignal_chunk(block_index=0,
                                             seg_index=0,
                                             i_start=0,
                                             i_stop=1024,
                                             stream_index=0,
                                             channel_indexes=range(10))
    >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk,
                                                         dtype='float64',
                                                         channel_indexes=[0, 3, 6])
    """

    extensions = ["edf"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        if not HAS_PYEDF:
            raise ValueError("Requires pyedflib")
        BaseRawIO.__init__(self)

        # note that this filename is used in self._source_name
        self.filename = filename

        self.signal_headers = []
        self.edf_header = {}

    def _source_name(self):
        return self.filename

    def _parse_header(self):

        # read basic header
        with open(self.filename, "rb") as f:
            f.seek(192)
            file_version_header = f.read(44).decode("ascii")
            # only accepting basic EDF files (no 'EDF+' in header)
            # or continuous EDF+ files ('EDF+C' in header)
            if ("EDF+" in file_version_header) and ("EDF+C" not in file_version_header):
                raise ValueError("Only continuous EDF+ files are currently supported.")

        self.edf_reader = EdfReader(self.filename)
        # load headers, signal information and
        self.edf_header = self.edf_reader.getHeader()
        self.signal_headers = self.edf_reader.getSignalHeaders()

        # add annotations to header
        annotations = self.edf_reader.readAnnotations()
        self.signal_annotations = [[s, d, a] for s, d, a in zip(*annotations)]

        # 1 stream = 1 sampling rate
        stream_characteristics = []
        self.stream_idx_to_chidx = {}

        signal_channels = []
        for ch_idx, sig_dict in enumerate(self.signal_headers):
            ch_name = sig_dict["label"]
            chan_id = ch_idx
            sr = sig_dict["sample_rate"]  # Hz
            dtype = "int16"  # assume general int16 based on edf documentation
            units = sig_dict["dimension"]
            physical_range = sig_dict["physical_max"] - sig_dict["physical_min"]
            # number of digital steps resolved (+1 to account for '0')
            digital_range = sig_dict["digital_max"] - sig_dict["digital_min"] + 1
            gain = physical_range / digital_range
            offset = -1 * sig_dict["digital_min"] * gain + sig_dict["physical_min"]

            # identify corresponding stream based on sampling rate
            if (sr,) not in stream_characteristics:
                stream_characteristics += [(sr,)]

            stream_id = stream_characteristics.index((sr,))
            self.stream_idx_to_chidx.setdefault(stream_id, []).append(ch_idx)

            signal_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))

        # convert channel index lists to arrays for indexing
        self.stream_idx_to_chidx = {k: np.array(v) for k, v in self.stream_idx_to_chidx.items()}

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        signal_streams = [(f"stream ({sr} Hz)", i) for i, sr in enumerate(stream_characteristics)]
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # no unit/epoch information contained in edf
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_channels.append(("Event", "event_channel", "event"))
        event_channels.append(("Epoch", "epoch_channel", "epoch"))
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]  # we only accept continuous edf files
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

        # Add custom annotations for neo objects
        bl_ann = self.raw_annotations["blocks"][0]
        bl_ann["name"] = "EDF Data Block"
        bl_ann.update(self.edf_header)
        seg_ann = bl_ann["segments"][0]
        seg_ann["name"] = "Seg #0 Block #0"

        # extract keys for array_annotations common to all signals and not already used
        ignore_annotations = [
            "label",
            "dimension",
            "sample_rate",
            "physical_min",
            "physical_max",
            "digital_min",
            "digital_max",
        ]
        array_keys = []
        for k in self.signal_headers[0]:
            if k not in ignore_annotations and all([k in h for h in self.signal_headers]):
                array_keys.append(k)

        for array_key in array_keys:
            array_anno = {array_key: [h[array_key] for h in self.signal_headers]}
        seg_ann["signals"].append({"__array_annotations__": array_anno})

    def _get_stream_channels(self, stream_index):
        return self.header["signal_channels"][self.stream_idx_to_chidx[stream_index]]

    def _segment_t_start(self, block_index, seg_index):
        # no time offset provided by EDF format
        return 0.0  # in seconds

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self.edf_reader.datarecord_duration * self.edf_reader.datarecords_in_file
        # this must return an float scale in second
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        chidx = self.stream_idx_to_chidx[stream_index][0]
        # use sample count of first signal in stream
        return self.edf_reader.getNSamples()[chidx]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0  # EDF does not provide temporal offset information

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        # only dealing with single block and segment edf files
        if block_index != 0:
            raise ValueError("`block_index` must be 0")
        if seg_index != 0:
            raise ValueError("`seg_index` must be 0")

        stream_channel_idxs = self.stream_idx_to_chidx[stream_index]

        # keep all channels of the stream if none are selected
        if channel_indexes is None:
            channel_indexes = slice(None)

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self.get_signal_size(block_index=block_index, seg_index=seg_index, stream_index=stream_index)
        n = i_stop - i_start

        # raw_signals = self.edf_reader. am[channel_indexes, i_start:i_stop]
        selected_channel_idxs = stream_channel_idxs[channel_indexes]

        # load data into numpy array buffer
        data = []
        for i, channel_idx in enumerate(selected_channel_idxs):
            # use int32 for compatibility with pyedflib
            buffer = np.empty(n, dtype=np.int32)
            self.edf_reader.read_digital_signal(channel_idx, i_start, n, buffer)
            data.append(buffer)

        # downgrade to int16 as this is what is used in the edf file format
        # use fortran (column major) order to be more efficient after transposing
        data = np.asarray(data, dtype=np.int16, order="F")

        # use dimensions (time, channel)
        data = data.T

        return data

    def _spike_count(self, block_index, seg_index, spike_channel_index):
        return None

    def _get_spike_timestamps(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        return None

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        return None

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        return None

    def _event_count(self, block_index, seg_index, event_channel_index):
        return len(self.edf_reader.readAnnotations()[0])

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        # these time should be already in seconds
        timestamps, durations, labels = self.edf_reader.readAnnotations()
        if t_start is None:
            t_start = self.segment_t_start(block_index, seg_index)
        if t_stop is None:
            t_stop = self.segment_t_stop(block_index, seg_index)

        # only consider events and epochs that overlap with t_start t_stop range
        time_mask = ((t_start < timestamps) & (timestamps < t_stop)) | (
            (t_start < (timestamps + durations)) & ((timestamps + durations) < t_stop)
        )

        # separate event from epoch times
        event_mask = durations[time_mask] == 0
        if self.header["event_channels"]["type"][event_channel_index] == b"epoch":
            event_mask = ~event_mask
            durations = durations[time_mask][event_mask]
        elif self.header["event_channels"]["type"][event_channel_index] == b"event":
            durations = None

        times = timestamps[time_mask][event_mask]
        labels = np.asarray(labels[time_mask][event_mask], dtype="U")

        return times, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        return np.asarray(event_timestamps, dtype=dtype)

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        return np.asarray(raw_duration, dtype=dtype)

    def __enter__(self):
        return self

    def __del__(self):
        self._close_reader()

    def __exit__(self, exc_type, exc_val, ex_tb):
        self._close_reader()

    def close(self):
        """
        Closes the file handler
        """
        self._close_reader()

    def _close_reader(self):
        if hasattr(self, "edf_reader"):
            self.edf_reader.close()



"""
Tests of neo.io.winwcpio
"""

import unittest

from neo.io import WinWcpIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestRawBinarySignalIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = WinWcpIO
    entities_to_download = ["winwcp"]
    entities_to_test = ["winwcp/File_winwcp_1.wcp"]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.plexon2io
"""

import unittest

from neo.io import Plexon2IO
from neo.test.iotest.common_io_test import BaseTestIO

from neo.test.rawiotest.test_plexon2rawio import TestPlexon2RawIO


try:
    from neo.rawio.plexon2rawio.pypl2 import pypl2lib

    HAVE_PYPL2 = True
except (ImportError, TimeoutError):
    HAVE_PYPL2 = False


@unittest.skipUnless(HAVE_PYPL2, "requires pypl package and all its dependencies")
class TestPlexon2IO(BaseTestIO, unittest.TestCase):
    entities_to_download = TestPlexon2RawIO.entities_to_download
    entities_to_test = TestPlexon2RawIO.entities_to_test
    ioclass = Plexon2IO


if __name__ == "__main__":
    unittest.main()



"""

"""

import unittest

import quantities as pq

from neo.io import OpenEphysBinaryIO
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.rawiotest.test_openephysbinaryrawio import TestOpenEphysBinaryRawIO


class TestOpenEphysBinaryIO(BaseTestIO, unittest.TestCase):
    ioclass = OpenEphysBinaryIO
    entities_to_download = TestOpenEphysBinaryRawIO.entities_to_download
    entities_to_test = TestOpenEphysBinaryRawIO.entities_to_test


if __name__ == "__main__":
    unittest.main()



"""
Tests of the neo.io.pickleio.PickleIO class
"""

import os

import unittest

import numpy as np
import quantities as pq

from neo.core import Block, Segment, AnalogSignal, SpikeTrain, Epoch, Event, IrregularlySampledSignal, Group
from neo.io import PickleIO
from numpy.testing import assert_array_equal
from neo.test.tools import assert_arrays_equal, assert_file_contents_equal
from neo.test.iotest.common_io_test import BaseTestIO

NCELLS = 5


class CommonTestPickleIO(BaseTestIO, unittest.TestCase):
    ioclass = PickleIO


class TestPickleIO(unittest.TestCase):
    def test__issue_285(self):
        # Spiketrain
        train = SpikeTrain([3, 4, 5] * pq.s, t_stop=10.0)
        unit = Group()
        unit.add(train)

        epoch = Epoch(np.array([0, 10, 20]), np.array([2, 2, 2]), np.array(["a", "b", "c"]), units="ms")

        blk = Block()
        seg = Segment()
        seg.spiketrains.append(train)
        seg.epochs.append(epoch)
        blk.segments.append(seg)

        reader = PickleIO(filename="blk.pkl")
        reader.write(blk)

        reader = PickleIO(filename="blk.pkl")
        r_blk = reader.read_block()
        r_seg = r_blk.segments[0]

        self.assertIsInstance(r_seg.epochs[0], Epoch)
        os.remove("blk.pkl")

        # Epoch
        epoch = Epoch(
            times=np.arange(0, 30, 10) * pq.s,
            durations=[10, 5, 7] * pq.ms,
            labels=np.array(["btn0", "btn1", "btn2"], dtype="U"),
        )
        blk = Block()
        seg = Segment()
        seg.epochs.append(epoch)
        blk.segments.append(seg)

        reader = PickleIO(filename="blk.pkl")
        reader.write(blk)

        reader = PickleIO(filename="blk.pkl")
        r_blk = reader.read_block()
        r_seg = r_blk.segments[0]
        self.assertIsInstance(r_seg.epochs[0].segment, Segment)
        os.remove("blk.pkl")

        # Event
        event = Event(np.arange(0, 30, 10) * pq.s, labels=np.array(["trig0", "trig1", "trig2"], dtype="U"))
        blk = Block()
        seg = Segment()
        seg.events.append(event)
        blk.segments.append(seg)

        reader = PickleIO(filename="blk.pkl")
        reader.write(blk)

        reader = PickleIO(filename="blk.pkl")
        r_blk = reader.read_block()
        r_seg = r_blk.segments[0]
        self.assertIsInstance(r_seg.events[0].segment, Segment)
        os.remove("blk.pkl")

        # IrregularlySampledSignal
        signal = IrregularlySampledSignal([0.0, 1.23, 6.78], [1, 2, 3], units="mV", time_units="ms")
        blk = Block()
        seg = Segment()
        seg.irregularlysampledsignals.append(signal)
        blk.segments.append(seg)

        reader = PickleIO(filename="blk.pkl")
        reader.write(blk)

        reader = PickleIO(filename="blk.pkl")
        r_blk = reader.read_block()
        r_seg = r_blk.segments[0]
        self.assertIsInstance(r_seg.irregularlysampledsignals[0].segment, Segment)
        os.remove("blk.pkl")


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.biocamrawio import BiocamRawIO


class BiocamIO(BiocamRawIO, BaseFromRaw):
    __doc__ = BiocamRawIO.__doc__
    mode = "file"

    def __init__(self, filename):
        BiocamRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Tests of neo.rawio.mearecrawio

"""

import unittest

from neo.rawio.mearecrawio import MEArecRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


try:
    import MEArec as mr

    HAVE_MEAREC = True
except ImportError:
    HAVE_MEAREC = False


@unittest.skipUnless(HAVE_MEAREC, "requires MEArec package")
class TestMEArecRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = MEArecRawIO
    entities_to_download = ["mearec"]
    entities_to_test = ["mearec/mearec_test_10s.h5"]

    def test_not_loading_recordings(self):

        filename = self.entities_to_test[0]
        filename = self.get_local_path(filename)
        rawio = self.rawioclass(filename=filename, load_analogsignal=False)
        rawio.parse_header()

        # Test that rawio does not have a _recordings attribute
        self.assertFalse(hasattr(rawio, "_recordings"))

        # Test that calling get_spike_timestamps works
        rawio.get_spike_timestamps()

        # Test that caling anlogsignal chunk raises the right error
        with self.assertRaises(AttributeError):
            rawio.get_analogsignal_chunk()

    def test_not_loading_spiketrain(self):

        filename = self.entities_to_test[0]
        filename = self.get_local_path(filename)
        rawio = self.rawioclass(filename=filename, load_spiketrains=False)
        rawio.parse_header()

        # Test that rawio does not have a _spiketrains attribute
        self.assertFalse(hasattr(rawio, "_spiketrains"))

        # Test that calling analogsignal chunk works
        rawio.get_analogsignal_chunk()

        # Test that calling get_spike_timestamps raises an the right error
        with self.assertRaises(AttributeError):
            rawio.get_spike_timestamps()


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.edfio
"""

import unittest
import numpy as np
import quantities as pq

from neo.io.edfio import EDFIO
from neo.test.iotest.common_io_test import BaseTestIO
from neo.io.proxyobjects import AnalogSignalProxy
from neo import AnalogSignal


class TestEDFIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = EDFIO
    entities_to_download = ["edf"]
    entities_to_test = [
        "edf/edf+C.edf",
    ]

    def setUp(self):
        super().setUp()
        self.filename = self.get_local_path("edf/edf+C.edf")

    def test_read_block(self):
        """
        Test reading the complete block and general annotations
        """
        with EDFIO(self.filename) as io:
            bl = io.read_block()
            self.assertTrue(bl.annotations)

            seg = bl.segments[0]
            assert seg.name == "Seg #0 Block #0"
            for anasig in seg.analogsignals:
                assert anasig.name is not None

    def test_read_segment_with_time_slice(self):
        """
        Test loading of a time slice and check resulting times
        """
        with EDFIO(self.filename) as io:
            seg = io.read_segment(time_slice=None)

            # data file does not contain spike, event or epoch timestamps
            self.assertEqual(len(seg.spiketrains), 0)
            self.assertEqual(len(seg.events), 1)
            self.assertEqual(len(seg.events[0]), 0)
            self.assertEqual(len(seg.epochs), 1)
            self.assertEqual(len(seg.epochs[0]), 0)
            for asig in seg.analogsignals:
                self.assertEqual(asig.shape[0], 256)
            n_channels = sum(a.shape[-1] for a in seg.analogsignals)
            self.assertEqual(n_channels, 5)

            t_start, t_stop = 500 * pq.ms, 800 * pq.ms
            seg = io.read_segment(time_slice=(t_start, t_stop))

            self.assertAlmostEqual(seg.t_start.rescale(t_start.units), t_start, delta=5.0)
            self.assertAlmostEqual(seg.t_stop.rescale(t_stop.units), t_stop, delta=5.0)

    def test_compare_data(self):
        """
        Compare data from AnalogSignal with plain data stored in text file
        """
        with EDFIO(self.filename) as io:
            plain_data = np.loadtxt(io.filename.replace(".edf", ".txt"), dtype=np.int16)
            seg = io.read_segment(lazy=True)

            anasigs = seg.analogsignals
            self.assertEqual(len(anasigs), 5)  # all channels have different units, so expecting 5
            for aidx, anasig in enumerate(anasigs):
                # comparing raw data to original values
                ana_data = anasig.load(magnitude_mode="raw")
                np.testing.assert_array_equal(ana_data.magnitude, plain_data[:, aidx : aidx + 1])

                # comparing floating data to original values * gain factor
                ch_head = io.edf_reader.getSignalHeader(aidx)
                physical_range = ch_head["physical_max"] - ch_head["physical_min"]
                # number of digital values used (+1 to account for '0' value)
                digital_range = ch_head["digital_max"] - ch_head["digital_min"] + 1

                gain = physical_range / digital_range
                ana_data = anasig.load(magnitude_mode="rescaled")
                rescaled_data = plain_data[:, aidx : aidx + 1] * gain
                np.testing.assert_array_equal(ana_data.magnitude, rescaled_data)


if __name__ == "__main__":
    unittest.main()



"""
Classe for reading/writing SpikeTrains in a text file.
It is the simple case where different spiketrains are written line by line.

Supported : Read/Write

Author: sgarcia

"""

import os

import numpy as np
import quantities as pq

from neo.io.baseio import BaseIO
from neo.core import Segment, SpikeTrain, NeoReadWriteError


class AsciiSpikeTrainIO(BaseIO):
    """

    Class for reading/writing SpikeTrains in a text file.
    Each Spiketrain is a line.

    Usage:
        >>> from neo import io
        >>> r = io.AsciiSpikeTrainIO( filename = 'File_ascii_spiketrain_1.txt')
        >>> seg = r.read_segment()
        >>> print seg.spiketrains     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        [<SpikeTrain(array([ 3.89981604,  4.73258781,  0.608428  ,  4.60246277,  1.23805797,
        ...

    """

    is_readable = True
    is_writable = True

    supported_objects = [Segment, SpikeTrain]
    readable_objects = [Segment]
    writeable_objects = [Segment]

    has_header = False
    is_streameable = False

    read_params = {
        Segment: [
            ("delimiter", {"value": "\t", "possible": ["\t", " ", ",", ";"]}),
            (
                "t_start",
                {
                    "value": 0.0,
                },
            ),
        ]
    }
    write_params = {
        Segment: [
            ("delimiter", {"value": "\t", "possible": ["\t", " ", ",", ";"]}),
        ]
    }

    name = None
    extensions = ["txt"]

    mode = "file"

    def __init__(self, filename=None):
        """
        This class read/write SpikeTrains in a text file.
        Each row is a spiketrain.

        **Arguments**

        filename : the filename to read/write

        """
        BaseIO.__init__(self)
        self.filename = filename

    def read_segment(
        self,
        lazy=False,
        delimiter="\t",
        t_start=0.0 * pq.s,
        unit=pq.s,
    ):
        """
        Arguments:
            delimiter  :  columns delimiter in file  '\t' or one space or two space or ',' or ';'
            t_start : time start of all spiketrain 0 by default
            unit : unit of spike times, can be a str or directly a Quantities
        """
        if lazy:
            raise NeoReadWriteError("This IO does not support lazy reading")

        unit = pq.Quantity(1, unit)

        seg = Segment(file_origin=os.path.basename(self.filename))

        with open(self.filename, "r", newline=None) as f:
            for i, line in enumerate(f):
                alldata = line[:-1].split(delimiter)
                if alldata[-1] == "":
                    alldata = alldata[:-1]
                if alldata[0] == "":
                    alldata = alldata[1:]

                spike_times = np.array(alldata).astype("f")
                t_stop = spike_times.max() * unit

                sptr = SpikeTrain(spike_times * unit, t_start=t_start, t_stop=t_stop)

                sptr.annotate(channel_index=i)
                seg.spiketrains.append(sptr)

        seg.check_relationships()
        return seg

    def write_segment(
        self,
        segment,
        delimiter="\t",
    ):
        """
        Write SpikeTrain of a Segment in a txt file.
        Each row is a spiketrain.

         Arguments:
            segment : the segment to write. Only analog signals will be written.
            delimiter  :  columns delimiter in file  '\t' or one space or two space or ',' or ';'

            information of t_start is lost

        """

        f = open(self.filename, "w")
        for s, sptr in enumerate(segment.spiketrains):
            for ts in sptr:
                f.write(f"{ts}{delimiter}")
            f.write("\n")
        f.close()



"""
neo.io has been split into a 2-level API:
  * neo.io: this API gives Neo objects
  * neo.rawio: this API gives raw data as they are in files.

Developers are encourage to use neo.rawio.

When this is done the neo.io can be implemented trivially
using code like shown in this file.

Author: sgarcia

"""

from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.examplerawio import ExampleRawIO


class ExampleIO(ExampleRawIO, BaseFromRaw):
    name = "example IO"
    description = "Fake IO"

    # This is an inportant choice when there are several channels.
    #   'split-all' :  1 AnalogSignal each 1 channel
    #   'group-by-same-units' : one 2D AnalogSignal for each group of channel with same units
    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename=""):
        ExampleRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



import unittest

from neo.io import RawMCSIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestRawMcsIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = RawMCSIO
    entities_to_download = ["rawmcs"]
    entities_to_test = ["rawmcs/raw_mcs_with_header_1.raw"]


if __name__ == "__main__":
    unittest.main()



"""
IO for reading edf and edf+ files using pyedflib

PyEDFLib
https://pyedflib.readthedocs.io
https://github.com/holgern/pyedflib

EDF Format Specifications: https://www.edfplus.info/

Author: Julia Sprenger
"""

from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.edfrawio import EDFRawIO


class EDFIO(EDFRawIO, BaseFromRaw):
    """
    IO for reading edf and edf+ files.
    """

    name = "EDF IO"
    description = "IO for reading EDF and EDF+ files"

    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename=""):
        EDFRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Test to  make sure generated datasets are sane.
"""

import unittest

from datetime import datetime

import numpy as np
import quantities as pq

from neo.core import class_by_name, Block, Segment, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch
from neo.core.dataobject import DataObject
from neo.test.generate_datasets import (
    generate_one_simple_block,
    generate_one_simple_segment,
    generate_from_supported_objects,
    TEST_ANNOTATIONS,
)
from neo.test.tools import assert_arrays_equal, assert_neo_object_is_compliant


class Test__generate_one_simple_segment(unittest.TestCase):
    def test_defaults(self):
        res = generate_one_simple_segment()

        self.assertTrue(isinstance(res, Segment))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.analogsignals), 0)
        self.assertEqual(len(res.irregularlysampledsignals), 0)
        self.assertEqual(len(res.spiketrains), 0)
        self.assertEqual(len(res.events), 0)
        self.assertEqual(len(res.epochs), 0)

    def test_all_supported(self):
        objects = [Block, Segment, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        res = generate_one_simple_segment(supported_objects=objects)

        self.assertTrue(isinstance(res, Segment))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.analogsignals), 4)
        self.assertEqual(len(res.irregularlysampledsignals), 0)
        self.assertEqual(len(res.spiketrains), 6)
        self.assertEqual(len(res.events), 3)
        self.assertEqual(len(res.epochs), 2)

    def test_half_supported(self):
        objects = [Segment, IrregularlySampledSignal, SpikeTrain, Epoch]

        res = generate_one_simple_segment(supported_objects=objects)

        self.assertTrue(isinstance(res, Segment))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.analogsignals), 0)
        self.assertEqual(len(res.irregularlysampledsignals), 0)
        self.assertEqual(len(res.spiketrains), 6)
        self.assertEqual(len(res.events), 0)
        self.assertEqual(len(res.epochs), 2)

    def test_all_without_block(self):
        objects = [Segment, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        res = generate_one_simple_segment(supported_objects=objects)

        self.assertTrue(isinstance(res, Segment))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.analogsignals), 4)
        self.assertEqual(len(res.irregularlysampledsignals), 0)
        self.assertEqual(len(res.spiketrains), 6)
        self.assertEqual(len(res.events), 3)
        self.assertEqual(len(res.epochs), 2)

    def test_all_without_segment_valueerror(self):
        objects = [Block, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        self.assertRaises(ValueError, generate_one_simple_segment, supported_objects=objects)


class Test__generate_one_simple_block(unittest.TestCase):
    def test_defaults(self):
        res = generate_one_simple_block()

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 0)

    def test_all_supported(self):
        objects = [Block, Segment, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        res = generate_one_simple_block(supported_objects=objects)

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 3)
        seg1, seg2, seg3 = res.segments

        self.assertEqual(len(seg1.analogsignals), 4)
        self.assertEqual(len(seg1.irregularlysampledsignals), 0)
        self.assertEqual(len(seg1.spiketrains), 6)
        self.assertEqual(len(seg1.events), 3)
        self.assertEqual(len(seg1.epochs), 2)

        self.assertEqual(len(seg2.analogsignals), 4)
        self.assertEqual(len(seg2.irregularlysampledsignals), 0)
        self.assertEqual(len(seg2.spiketrains), 6)
        self.assertEqual(len(seg2.events), 3)
        self.assertEqual(len(seg2.epochs), 2)

        self.assertEqual(len(seg3.analogsignals), 4)
        self.assertEqual(len(seg3.irregularlysampledsignals), 0)
        self.assertEqual(len(seg3.spiketrains), 6)
        self.assertEqual(len(seg3.events), 3)
        self.assertEqual(len(seg3.epochs), 2)

    def test_half_supported(self):
        objects = [Block, Segment, IrregularlySampledSignal, SpikeTrain, Epoch]

        res = generate_one_simple_block(supported_objects=objects)

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 3)
        seg1, seg2, seg3 = res.segments

        self.assertEqual(len(seg1.analogsignals), 0)
        self.assertEqual(len(seg1.irregularlysampledsignals), 0)
        self.assertEqual(len(seg1.spiketrains), 6)
        self.assertEqual(len(seg1.events), 0)
        self.assertEqual(len(seg1.epochs), 2)

        self.assertEqual(len(seg2.analogsignals), 0)
        self.assertEqual(len(seg2.irregularlysampledsignals), 0)
        self.assertEqual(len(seg2.spiketrains), 6)
        self.assertEqual(len(seg2.events), 0)
        self.assertEqual(len(seg2.epochs), 2)

        self.assertEqual(len(seg3.analogsignals), 0)
        self.assertEqual(len(seg3.irregularlysampledsignals), 0)
        self.assertEqual(len(seg3.spiketrains), 6)
        self.assertEqual(len(seg3.events), 0)
        self.assertEqual(len(seg3.epochs), 2)

    def test_all_without_segment(self):
        objects = [Block, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        res = generate_one_simple_block(supported_objects=objects)

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 0)

    def test_all_without_block_valueerror(self):
        objects = [Segment, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        self.assertRaises(ValueError, generate_one_simple_block, supported_objects=objects)


class Test__generate_from_supported_objects(unittest.TestCase):
    def test_no_object_valueerror(self):
        objects = []

        self.assertRaises(ValueError, generate_from_supported_objects, objects)

    def test_all(self):
        objects = [Block, Segment, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        res = generate_from_supported_objects(objects)

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 3)
        seg1, seg2, seg3 = res.segments

        self.assertEqual(len(seg1.analogsignals), 4)
        self.assertEqual(len(seg1.irregularlysampledsignals), 0)
        self.assertEqual(len(seg1.spiketrains), 6)
        self.assertEqual(len(seg1.events), 3)
        self.assertEqual(len(seg1.epochs), 2)

        self.assertEqual(len(seg2.analogsignals), 4)
        self.assertEqual(len(seg2.irregularlysampledsignals), 0)
        self.assertEqual(len(seg2.spiketrains), 6)
        self.assertEqual(len(seg2.events), 3)
        self.assertEqual(len(seg2.epochs), 2)

        self.assertEqual(len(seg3.analogsignals), 4)
        self.assertEqual(len(seg3.irregularlysampledsignals), 0)
        self.assertEqual(len(seg3.spiketrains), 6)
        self.assertEqual(len(seg3.events), 3)
        self.assertEqual(len(seg3.epochs), 2)

    def test_block(self):
        objects = [Block]

        res = generate_from_supported_objects(objects)

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 0)

    def test_block_segment(self):
        objects = [Segment, Block]

        res = generate_from_supported_objects(objects)

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 3)
        seg1, seg2, seg3 = res.segments

        self.assertEqual(len(seg1.analogsignals), 0)
        self.assertEqual(len(seg1.irregularlysampledsignals), 0)
        self.assertEqual(len(seg1.spiketrains), 0)
        self.assertEqual(len(seg1.events), 0)
        self.assertEqual(len(seg1.epochs), 0)

        self.assertEqual(len(seg2.analogsignals), 0)
        self.assertEqual(len(seg2.irregularlysampledsignals), 0)
        self.assertEqual(len(seg2.spiketrains), 0)
        self.assertEqual(len(seg2.events), 0)
        self.assertEqual(len(seg2.epochs), 0)

        self.assertEqual(len(seg3.analogsignals), 0)
        self.assertEqual(len(seg3.irregularlysampledsignals), 0)
        self.assertEqual(len(seg3.spiketrains), 0)
        self.assertEqual(len(seg3.events), 0)
        self.assertEqual(len(seg3.epochs), 0)

    def test_segment(self):
        objects = [Segment]

        res = generate_from_supported_objects(objects)

        self.assertTrue(isinstance(res, Segment))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.analogsignals), 0)
        self.assertEqual(len(res.irregularlysampledsignals), 0)
        self.assertEqual(len(res.spiketrains), 0)
        self.assertEqual(len(res.events), 0)
        self.assertEqual(len(res.epochs), 0)

    def test_all_without_block(self):
        objects = [Segment, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        res = generate_from_supported_objects(objects)

        self.assertTrue(isinstance(res, Segment))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.analogsignals), 4)
        self.assertEqual(len(res.irregularlysampledsignals), 0)
        self.assertEqual(len(res.spiketrains), 6)
        self.assertEqual(len(res.events), 3)
        self.assertEqual(len(res.epochs), 2)

    def test_all_without_segment(self):
        objects = [Block, AnalogSignal, IrregularlySampledSignal, SpikeTrain, Event, Epoch]

        res = generate_from_supported_objects(supported_objects=objects)

        self.assertTrue(isinstance(res, Block))
        assert_neo_object_is_compliant(res)

        self.assertEqual(len(res.segments), 0)



"""
Objects to represent continguous sections of samples in a Neuralynx continuously sampled .Ncs
file. These sections are considered contiguous in that they have a start time, a sampling rate,
and a length.

Defining these sections is complicated due to the physical structure of .Ncs files which contain
both a header and a set of fixed length records, CscRecords.

Each CscRecord has a start time in microseconds, a channel id (chan_id), a stated sampling
frequency (sampFreq) which is rounded to the whole sample/s, and a number of valid samples
(nb_valid), which may be less than the physical maximum of 512. In principle each of these
parameters may vary in each record in the file; however, there are no known examples of .Ncs
files where the chan_id or sampFreq varied from record to record.

The header normally, though not always, contains a stated sampling frequency, which may be
rounded to a whole number of samples per second or not.

Finally, the actual sampling frequency used within a section may be slightly different than any of
the two frequencies above (that in the records or that in the header). This often arises due to
clock drift over the course of a longer recording combined with the fact that only whole
microseconds are recorded in the CscRecords. This can easily be 0.01% amounting to 0.2 seconds over
the course of a half hour recording.

These files may often contain apparent gaps of time between the sequential CscRecords in a file.
A gap is where the predicted time at the start of the next record, based on taking the start time
of the record and then adding 1/sampling rate x nbValid, does not agree exactly with the start time
given in the next record.

These gaps vary from those of less than one microsecond, likely due to rounding of the start time,
to gaps of many minutes, which may be introduced by the human operator stopping and starting
recording during a recording session.

The challenge addressed by the NcsSectionsFactory of this module is to separate those intended
gaps from those introduced by a quirk of the hardware, recording software, or file format.
"""

import math
import numpy as np


class NcsSections:
    """
    Contains information regarding the contiguous sections of records in an Ncs file.
    Methods of NcsSectionsFactory perform parsing of this information from an Ncs file and
    produce these where the sections are discontiguous in time and in temporal order.

    TODO: This class will likely need __ne__ to be useful in
    more sophisticated segment construction algorithms.

    """

    def __init__(self):
        self.sects = []
        self.sampFreqUsed = 0  # actual sampling frequency of samples
        self.microsPerSampUsed = 0  # microseconds per sample

    def __eq__(self, other):
        samp_eq = self.sampFreqUsed == other.sampFreqUsed
        micros_eq = self.microsPerSampUsed == other.microsPerSampUsed
        sects_eq = self.sects == other.sects
        return samp_eq and micros_eq and sects_eq

    def __hash__(self):
        return (f"{self.sampFreqUsed};{self.microsPerSampUsed};" f"{[s.__hash__() for s in self.sects]}").__hash__()

    def is_equivalent(self, other, rel_tol=0, abs_tol=0):
        if len(self.sects) != len(other.sects):
            return False
        else:
            # do not check for gaps if only a single section is present
            for sec_id in range(len(self.sects) - 1):
                if not self.sects[sec_id].is_equivalent(other.sects[sec_id], rel_tol=rel_tol, abs_tol=abs_tol):
                    return False
            return True


class NcsSection:
    """
    Information regarding a single contiguous section or group of records in an Ncs file.

    Model is that times are closed on the left and open on the right. Record
    numbers are closed on both left and right, that is, inclusive of the last record.

    endTime should never be set less than startTime for comparison functions to work
    properly, though this is not enforced.
    """

    _RECORD_SIZE = 512  # nb sample per signal record

    def __init__(self, startRec, startTime, endRec, endTime, n_samples):
        self.startRec = startRec  # index of starting record
        self.startTime = startTime  # starttime of first record
        self.endRec = endRec  # index of last record (inclusive)
        self.endTime = endTime  # end time of last record, that is, the end time of the last
        # sampling period contained in the last record of the section
        self.n_samples = n_samples  # number of samples in record which are valid

    def __eq__(self, other):
        return (
            self.startRec == other.startRec
            and self.startTime == other.startTime
            and self.endRec == other.endRec
            and self.endTime == other.endTime
            and self.n_samples == other.n_samples
        )

    def __hash__(self):
        s = f"{self.startRec};{self.startTime};{self.endRec};{self.endTime};{self.n_samples}"
        return s.__hash__()

    def is_equivalent(self, other, rel_tol=0, abs_tol=0):
        eq_start = math.isclose(self.startTime, other.startTime, rel_tol=rel_tol, abs_tol=abs_tol)
        eq_end = math.isclose(self.endTime, other.endTime, rel_tol=rel_tol, abs_tol=abs_tol)
        return eq_start & eq_end

    def before_time(self, rhb):
        """
        Determine if this section is completely before another section in time.
        """
        return self.endTime < rhb.startTime

    def overlaps_time(self, rhb):
        """
        Determine if this section overlaps another in time.
        """
        return self.startTime <= rhb.endTime and self.endTime >= rhb.startTime

    def after_time(self, rhb):
        """
        Determine if this section is completely after another section in time.
        """
        return self.startTime >= rhb.endTime


class NcsSectionsFactory:
    """
    Class for factory methods which perform parsing of contiguous sections of records
    in Ncs files.

    Model for times is that times are rounded to nearest microsecond. Times
    from start of a sample until just before the next sample are included,
    that is, closed lower bound and open upper bound on intervals. A
    channel with no samples is empty and contains no time intervals.

    Moved here since algorithm covering all 3 header styles and types used is
    more complicated.
    """

    _maxGapSampFrac = 0.2  # maximum fraction of a sampling interval between predicted
    # and actual record timestamps still considered within one section

    @staticmethod
    def get_freq_for_micros_per_samp(micros):
        """
        Compute fractional sampling frequency, given microseconds per sample.
        """
        return 1e6 / micros

    @staticmethod
    def get_micros_per_samp_for_freq(sampFr):
        """
        Calculate fractional microseconds per sample, given the sampling frequency (Hz).
        """
        return 1e6 / sampFr

    @staticmethod
    def calc_sample_time(sampFr, startTime, posn):
        """
        Calculate time rounded to microseconds for sample given frequency,
        start time, and sample position.
        """
        return round(startTime + NcsSectionsFactory.get_micros_per_samp_for_freq(sampFr) * posn)

    @staticmethod
    def _buildNcsSections(ncsMemMap, sampFreq, gapTolerance=0):
        """
        Construct NcsSections with fast mode when no gaps or parsing the file to detect gaps.
        """
        channel_id = ncsMemMap["channel_id"][0]

        ncsSects = NcsSections()

        # check if file is one block of records, which is often the case, and avoid full parse
        # need that last timestamp match the predicted one.
        predLastBlockStartTime = NcsSectionsFactory.calc_sample_time(
            sampFreq, ncsMemMap["timestamp"][0], NcsSection._RECORD_SIZE * (ncsMemMap.shape[0] - 1)
        )
        if (
            channel_id == ncsMemMap["channel_id"][-1]
            and ncsMemMap["sample_rate"][0] == ncsMemMap["sample_rate"][-1]
            and ncsMemMap["timestamp"][-1] == predLastBlockStartTime
        ):
            lastBlkEndTime = NcsSectionsFactory.calc_sample_time(
                sampFreq, ncsMemMap["timestamp"][-1], ncsMemMap["nb_valid"][-1]
            )
            n_samples = NcsSection._RECORD_SIZE * (ncsMemMap.size - 1) + ncsMemMap["nb_valid"][-1]
            section0 = NcsSection(
                startRec=0,
                startTime=ncsMemMap["timestamp"][0],
                endRec=ncsMemMap.size - 1,
                endTime=lastBlkEndTime,
                n_samples=n_samples,
            )
            ncsSects.sects.append(section0)

        else:
            # need to parse all data block to detect gaps
            # check when the predicted timestamp is outside the tolerance
            delta = (ncsMemMap["timestamp"][1:] - ncsMemMap["timestamp"][:-1]).astype(np.int64)
            delta_prediction = ((ncsMemMap["nb_valid"][:-1] / sampFreq) * 1e6).astype(np.int64)

            gap_inds = np.flatnonzero(np.abs(delta - delta_prediction) > gapTolerance)
            gap_inds += 1

            sections_limits = [0] + gap_inds.tolist() + [len(ncsMemMap)]

            for i in range(len(gap_inds) + 1):
                start = sections_limits[i]
                stop = sections_limits[i + 1]
                duration = np.uint64(1e6 / sampFreq * ncsMemMap["nb_valid"][stop - 1])
                ncsSects.sects.append(
                    NcsSection(
                        startRec=start,
                        startTime=ncsMemMap["timestamp"][start],
                        endRec=stop - 1,
                        endTime=ncsMemMap["timestamp"][stop - 1] + duration,
                        n_samples=np.sum(ncsMemMap["nb_valid"][start:stop]),
                    )
                )

        return ncsSects

    @staticmethod
    def build_for_ncs_file(ncsMemMap, nlxHdr, gapTolerance=None, strict_gap_mode=True):
        """
        Build an NcsSections object for an NcsFile, given as a memmap and NlxHeader,
        handling gap detection appropriately given the file type as specified by the header.

        Parameters
        ----------
        ncsMemMap:
            memory map of file
        nlxHdr:
            NlxHeader from corresponding file.

        Returns
        -------
        An NcsSections corresponding to the provided ncsMemMap and nlxHdr
        """
        acqType = nlxHdr.type_of_recording()
        freq = nlxHdr["sampling_rate"]

        if acqType == "PRE4":
            # Old Neuralynx style with truncated whole microseconds for actual sampling. This
            # restriction arose from the sampling being based on a master 1 MHz clock.
            microsPerSampUsed = math.floor(NcsSectionsFactory.get_micros_per_samp_for_freq(freq))
            sampFreqUsed = NcsSectionsFactory.get_freq_for_micros_per_samp(microsPerSampUsed)
            if gapTolerance is None:
                if strict_gap_mode:
                    # this is the old behavior, maybe we could put 0.9 sample interval no ?
                    gapTolerance = 0
                else:
                    gapTolerance = 0

            ncsSects = NcsSectionsFactory._buildNcsSections(ncsMemMap, sampFreqUsed, gapTolerance=gapTolerance)
            ncsSects.sampFreqUsed = sampFreqUsed
            ncsSects.microsPerSampUsed = microsPerSampUsed

        elif acqType in ["DIGITALLYNX", "DIGITALLYNXSX", "CHEETAH64", "CHEETAH560", "RAWDATAFILE"]:
            # digital lynx style with fractional frequency and micros per samp determined from block times
            if gapTolerance is None:
                if strict_gap_mode:
                    # this is the old behavior
                    gapTolerance = round(NcsSectionsFactory._maxGapSampFrac * 1e6 / freq)
                else:
                    # quarter of paquet size is tolerate
                    gapTolerance = round(0.25 * NcsSection._RECORD_SIZE * 1e6 / freq)
            ncsSects = NcsSectionsFactory._buildNcsSections(ncsMemMap, freq, gapTolerance=gapTolerance)

            # take longer data block to compute reaal sampling rate
            # ind_max = np.argmax([section.n_samples for section in ncsSects.sects])
            ind_max = np.argmax([section.endRec - section.startRec for section in ncsSects.sects])
            section = ncsSects.sects[ind_max]
            if section.endRec != section.startRec:
                # need several data block
                sampFreqUsed = (
                    (section.n_samples - ncsMemMap["nb_valid"][section.endRec])
                    * 1e6
                    / (ncsMemMap["timestamp"][section.endRec] - section.startTime)
                )
            else:
                sampFreqUsed = freq
            ncsSects.sampFreqUsed = sampFreqUsed
            ncsSects.microsPerSampUsed = NcsSectionsFactory.get_micros_per_samp_for_freq(sampFreqUsed)

        elif acqType == "BML" or acqType == "ATLAS":
            # BML & ATLAS style with fractional frequency and micros per samp
            if strict_gap_mode:
                # this is the old behavior, maybe we could put 0.9 sample interval no ?
                gapTolerance = 0
            else:
                # quarter of paquet size is tolerate
                gapTolerance = round(0.25 * NcsSection._RECORD_SIZE * 1e6 / freq)
            ncsSects = NcsSectionsFactory._buildNcsSections(ncsMemMap, freq, gapTolerance=gapTolerance)
            ncsSects.sampFreqUsed = freq
            ncsSects.microsPerSampUsed = NcsSectionsFactory.get_micros_per_samp_for_freq(freq)

        else:
            raise TypeError("Unknown Ncs file type from header.")

        return ncsSects

    @staticmethod
    def _verifySectionsStructure(ncsMemMap, ncsSects):
        """
        Check that the record structure and timestamps for the ncsMemMap
        agrees with that in ncsSects.

        Provides a more rapid verification of structure than building a new NcsSections
        and checking equality.

        Parameters
        ----------
        ncsMemMap:
            memmap of file to be checked
        ncsSects
            existing block structure to be checked

        Returns
        -------
            true if all timestamps and block record starts and stops agree, otherwise false.
        """
        for blki in range(0, len(ncsSects.sects)):
            if ncsMemMap["timestamp"][ncsSects.sects[blki].startRec] != ncsSects.sects[blki].startTime:
                return False
            ets = ncsMemMap["timestamp"][ncsSects.sects[blki].endRec]
            enb = ncsMemMap["nb_valid"][ncsSects.sects[blki].endRec]
            endTime = NcsSectionsFactory.calc_sample_time(ncsSects.sampFreqUsed, ets, enb)
            if endTime != ncsSects.sects[blki].endTime:
                return False

        return True



import unittest
from platform import system
from sys import maxsize

try:
    if system() == "Windows":
        if maxsize > 2**32:
            import sonpy.amd64.sonpy
        else:
            import sonpy.win32.sonpy
    elif system() == "Darwin":
        import sonpy.darwin.sonpy
    elif system() == "Linux":
        import sonpy.linux.sonpy
    from neo.io import CedIO
except ImportError:
    HAVE_SONPY = False
    CedIO = None
else:
    HAVE_SONPY = True

from neo.test.iotest.common_io_test import BaseTestIO


@unittest.skipUnless(HAVE_SONPY, "sonpy")
class TestCedIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = CedIO
    entities_to_test = ["spike2/m365_1sec.smrx", "spike2/File_spike2_1.smr", "spike2/Two-mice-bigfile-test000.smr"]
    entities_to_download = ["spike2"]


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.blackrockrawio import BlackrockRawIO


class BlackrockIO(BlackrockRawIO, BaseFromRaw):
    """
    Supplementary class for reading BlackRock data using only a single nsx file.
    """

    name = "Blackrock IO for single nsx"
    description = (
        "This IO reads a pair of corresponding nev and nsX files of the Blackrock " "" + "(Cerebus) recording system."
    )

    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename, **kargs):
        BlackrockRawIO.__init__(self, filename=filename, **kargs)
        BaseFromRaw.__init__(self, filename)



"""

"""

import unittest

from neo.rawio.micromedrawio import MicromedRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestMicromedRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = MicromedRawIO
    entities_to_download = ["micromed"]
    entities_to_test = ["micromed/File_micromed_1.TRC"]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.neomatlabio
"""

import unittest

from neo.io import MicromedIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestMicromedIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = MicromedIO
    entities_to_download = ["micromed"]
    entities_to_test = ["micromed/File_micromed_1.TRC"]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.igorproio
"""

import unittest

try:
    import igor2

    HAVE_IGOR = True
except ImportError:
    HAVE_IGOR = False
from neo.io.igorproio import IgorIO
from neo.test.iotest.common_io_test import BaseTestIO


@unittest.skipUnless(HAVE_IGOR, "requires igor")
class TestIgorIO(BaseTestIO, unittest.TestCase):
    ioclass = IgorIO
    entities_to_download = ["igor"]
    entities_to_test = ["igor/mac-version2.ibw", "igor/win-version2.ibw"]


if __name__ == "__main__":
    unittest.main()



"""
Class for reading data from WinEdr, a software tool written by
John Dempster.

WinEdr is free:
http://spider.science.strath.ac.uk/sipbs/software.htm

Author: Samuel Garcia

"""

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
    _common_sig_characteristics,
)


class WinEdrRawIO(BaseRawIO):
    extensions = ["EDR", "edr"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        """
        Class for reading WinEdr data

        Parameters
        ----------
        filename: str, default: ''
            The *.edr file to be loaded

        """
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        with open(self.filename, "rb") as fid:
            headertext = fid.read(2048)
            headertext = headertext.decode("ascii")
            header = {}
            for line in headertext.split("\r\n"):
                if "=" not in line:
                    continue
                # print '#' , line , '#'
                key, val = line.split("=")
                if key in ["NC", "NR", "NBH", "NBA", "NBD", "ADCMAX", "NP", "NZ", "ADCMAX"]:
                    val = int(val)
                elif key in [
                    "AD",
                    "DT",
                ]:
                    val = val.replace(",", ".")
                    val = float(val)
                header[key] = val

        self._raw_signals = np.memmap(
            self.filename,
            dtype="int16",
            mode="r",
            shape=(
                header["NP"] // header["NC"],
                header["NC"],
            ),
            offset=header["NBH"],
        )

        DT = header["DT"]
        if "TU" in header:
            if header["TU"] == "ms":
                DT *= 0.001
        self._sampling_rate = 1.0 / DT

        signal_channels = []
        for c in range(header["NC"]):
            YCF = float(header[f"YCF{c}"].replace(",", "."))
            YAG = float(header[f"YAG{c}"].replace(",", "."))
            YZ = float(header[f"YZ{c}"].replace(",", "."))
            ADCMAX = header["ADCMAX"]
            AD = header["AD"]

            name = header[f"YN{c}"]
            chan_id = header[f"YO{c}"]
            units = header[f"YU{c}"]
            gain = AD / (YCF * YAG * (ADCMAX + 1))
            offset = -YZ * gain
            stream_id = "0"
            signal_channels.append((name, str(chan_id), self._sampling_rate, "int16", units, gain, offset, stream_id))

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        characteristics = signal_channels[_common_sig_characteristics]
        unique_characteristics = np.unique(characteristics)
        signal_streams = []
        for i in range(unique_characteristics.size):
            mask = unique_characteristics[i] == characteristics
            signal_channels["stream_id"][mask] = str(i)
            signal_streams.append((f"stream {i}", str(i)))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._raw_signals.shape[0] / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        return self._raw_signals.shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        (global_channel_indexes,) = np.nonzero(self.header["signal_channels"]["stream_id"] == stream_id)
        if channel_indexes is None:
            channel_indexes = slice(None)
        global_channel_indexes = global_channel_indexes[channel_indexes]
        raw_signals = self._raw_signals[slice(i_start, i_stop), global_channel_indexes]
        return raw_signals



"""
This module defines :class:`DataObject`, the abstract base class
used by all :module:`neo.core` classes that can contain data (i.e. are not container classes).
It contains basic functionality that is shared among all those data objects.

"""

from copy import deepcopy
import warnings

import quantities as pq
import numpy as np

from neo.core.baseneo import BaseNeo, _check_annotations


def _normalize_array_annotations(value, length):
    """Check consistency of array annotations

    Recursively check that value is either an array or list containing only "simple" types
    (number, string, date/time) or is a dict of those.

    Parameters
    ----------
    value : np.ndarray or list or tuple or dict
        Value to be checked for consistency.
    length : int
        Required length of the array annotation.

    Returns
    -------
    np.ndarray
        The array_annotations from value in correct form

    Raises
    ------
    ValueError
        In case value is not accepted as array_annotation(s)
    """

    # First stage, resolve dict of annotations into single annotations
    if isinstance(value, dict):
        for key in value.keys():
            if isinstance(value[key], dict):
                raise ValueError("Nested dicts are not allowed as array annotations")

            value[key] = _normalize_array_annotations(value[key], length)

    elif value is None:
        raise ValueError("Array annotations must not be None")
    # If not array annotation, pass on to regular check and make it a list, that is checked again
    # This covers array annotations with length 1
    elif not isinstance(value, (list, np.ndarray, tuple)) or (isinstance(value, pq.Quantity) and value.shape == ()):
        _check_annotations(value)
        value = _normalize_array_annotations(np.array([value]), length)

    # If array annotation, check for correct length, only single dimension and allowed data
    else:
        # Get length that is required for array annotations, which is equal to the length
        # of the object's data
        own_length = length

        # Escape check if empty array or list and just annotate an empty array (length 0)
        # This enables the user to easily create dummy array annotations that will be filled
        # with data later on
        if len(value) == 0:
            if not isinstance(value, np.ndarray):
                value = np.ndarray((0,))
            val_length = own_length
        else:
            # Note: len(o) also works for np.ndarray, it then uses the first dimension,
            # which is exactly the desired behaviour here
            val_length = len(value)

        if not own_length == val_length:
            raise ValueError(f"Incorrect length of array annotation: {val_length} != {own_length}")

        # Local function used to check single elements of a list or an array
        # They must not be lists or arrays and fit the usual annotation data types
        def _check_single_elem(element):
            # Nested array annotations not allowed currently
            # If element is a list or a np.ndarray, it's not conform except if it's a quantity of
            # length 1
            if isinstance(element, list) or (
                isinstance(element, np.ndarray)
                and not (isinstance(element, pq.Quantity) and (element.shape == () or element.shape == (1,)))
            ):
                raise ValueError("Array annotations should only be 1-dimensional")
            if isinstance(element, dict):
                raise ValueError("Dictionaries are not supported as array annotations")

            # Perform regular check for elements of array or list
            _check_annotations(element)

        # Arrays only need testing of single element to make sure the others are the same
        if isinstance(value, np.ndarray):
            # Type of first element is representative for all others
            # Thus just performing a check on the first element is enough
            # Even if it's a pq.Quantity, which can be scalar or array, this is still true
            # Because a np.ndarray cannot contain scalars and sequences simultaneously

            # If length of data is 0, then nothing needs to be checked
            if len(value):
                # Perform check on first element
                _check_single_elem(value[0])

            return value

        # In case of list, it needs to be ensured that all data are of the same type
        else:
            # Conversion to numpy array makes all elements same type
            # Converts elements to most general type

            try:
                value = np.array(value)
            # Except when scalar and non-scalar values are mixed, this causes conversion to fail
            except ValueError as e:
                msg = str(e)
                if "setting an array element with a sequence." in msg:
                    raise ValueError(
                        "Scalar values and arrays/lists cannot be " "combined into a single array annotation"
                    )
                else:
                    raise e

            # If most specialized data type that possibly fits all elements is object,
            # raise an Error with a telling error message, because this means the elements
            # are not compatible
            if value.dtype == object:
                raise ValueError("Cannot convert list of incompatible types into a single" " array annotation")

            # Check the first element for correctness
            # If its type is correct for annotations, all others are correct as well
            # Note: Empty lists cannot reach this point
            _check_single_elem(value[0])

    return value


class DataObject(BaseNeo, pq.Quantity):
    """
    This is the base class from which all objects containing data inherit
    It contains common functionality for all those objects and handles array_annotations.

    Parameters
    ----------
    name: str | None, default: None
        Name of the Neo object
    description: str | None, default: None
        Human readable string description of the Neo object
    array_annotations: dict | None, default: None
        Dictionary containing arrays / lists which annotate individual data points of the Neo object.
    **annotations: dict | None:
        Other keyword annotations to be included

    Notes
    -----

    Common functionality that is not included in BaseNeo includes:
    - duplicating with new data
    - rescaling the object
    - copying the object
    - returning it as pq.Quantity or np.ndarray
    - handling of array_annotations

    Array_annotations are a kind of annotation that contains metadata for every data point,
    i.e. per timestamp (in SpikeTrain, Event and Epoch) or signal channel (in AnalogSignal
    and IrregularlySampledSignal).
    They can contain the same data types as regular annotations, but are always represented
    as numpy arrays of the same length as the number of data points of the annotated neo object.

    """

    def __init__(self, name=None, description=None, file_origin=None, array_annotations=None, **annotations):
        """
        This method is called by each data object and initializes the newly created object by
        adding array annotations and calling __init__ of the super class, where more annotations
        and attributes are processed.
        """

        if not hasattr(self, "array_annotations") or not self.array_annotations:
            self.array_annotations = ArrayDict(self._get_arr_ann_length())
        if array_annotations is not None:
            self.array_annotate(**array_annotations)

        BaseNeo.__init__(self, name=name, description=description, file_origin=file_origin, **annotations)

    def array_annotate(self, **array_annotations):
        """
        Add array annotations (annotations for individual data points) as arrays to a Neo data
        object.

        Parameters
        ----------
        **array_annotations: dict
            Series of keyword annotations to add to the object

        Examples
        --------
        >>> obj.array_annotate(code=['a', 'b', 'a'], category=[2, 1, 1])
        >>> obj.array_annotations['code'][1]
        'b'
        """

        self.array_annotations.update(array_annotations)

    def array_annotations_at_index(self, index):
        """
        Return dictionary of array annotations at a given index or list of indices

        Parameters
        ----------
        index: int | list | np.ndarray
            The index (indices) from which the annotations are extracted

        Returns
        -------
        index_annotations: dict
            Dictionary of values or numpy arrays containing all array annotations for given index/indices

        Examples
        --------
        >>> obj.array_annotate(code=['a', 'b', 'a'], category=[2, 1, 1])
        >>> obj.array_annotations_at_index(1)
        {code='b', category=1}
        """

        # Taking only a part of the array annotations
        # Thus not using ArrayDict here, because checks for length are not needed
        index_annotations = {}

        # Use what is given as an index to determine the corresponding annotations,
        # if not possible, numpy raises an Error
        for ann in self.array_annotations.keys():
            # NO deepcopy, because someone might want to alter the actual object using this
            try:
                index_annotations[ann] = self.array_annotations[ann][index]
            except IndexError as e:
                # IndexError caused by 'dummy' array annotations should not result in failure
                # Taking a slice from nothing results in nothing
                if len(self.array_annotations[ann]) == 0 and not self._get_arr_ann_length() == 0:
                    index_annotations[ann] = self.array_annotations[ann]
                else:
                    raise e

        return index_annotations

    def _merge_array_annotations(self, other):
        """
        Merges array annotations of 2 different objects.

        Parameters
        ----------
        other: any
            The annotation to attemp to merge

        Returns
        -------
        merged_array_annotations: dict
            The merged annotations

        Notes
        -----
        The merge happens in such a way that the result fits the merged data
        In general this means concatenating the arrays from the 2 objects.
        If an annotation is only present in one of the objects, it will be omitted
        """

        merged_array_annotations = {}
        omitted_keys_self = []
        # Concatenating arrays for each key
        for key in self.array_annotations:
            try:
                value = deepcopy(self.array_annotations[key])
                other_value = deepcopy(other.array_annotations[key])
                # Quantities need to be rescaled to common unit
                if isinstance(value, pq.Quantity):
                    try:
                        other_value = other_value.rescale(value.units)
                    except ValueError:
                        raise ValueError("Could not merge array annotations " "due to different units")
                    merged_array_annotations[key] = np.append(value, other_value) * value.units
                else:
                    merged_array_annotations[key] = np.append(value, other_value)

            except KeyError:
                # Save the  omitted keys to be able to print them
                omitted_keys_self.append(key)
                continue
        # Also save omitted keys from 'other'
        omitted_keys_other = [key for key in other.array_annotations if key not in self.array_annotations]

        # Warn if keys were omitted
        if omitted_keys_other or omitted_keys_self:
            warnings.warn(
                f"The following array annotations were omitted, because they were only "
                f"present in one of the merged objects: {omitted_keys_self} from the "
                f"one that was merged into and {omitted_keys_other} from the one that "
                f"was merged into the other",
                UserWarning,
            )

        # Return the merged array_annotations
        return merged_array_annotations

    def rescale(self, units, dtype=None):
        """
        Return a copy of the object converted to the specified units.

        Parameters
        ----------
        units: quantity unit
            The units to convert the object to
        dtype: a numpy dtype
            Only exists for backward compatibility see [1]

        Returns
        -------
        self.copy(): Any
            A copy of the object with the desired units

        Notes
        -----
        [1] The `dtype` argument exists only for backward compatibility within quantities, see
        https://github.com/python-quantities/python-quantities/pull/204
        """

        # Use simpler functionality, if nothing will be changed
        dim = pq.quantity.validate_dimensionality(units)
        if self.dimensionality == dim:
            return self.copy()

        # Rescale the object into a new object
        obj = self.duplicate_with_new_data(signal=self.view(pq.Quantity).rescale(dim), units=units)

        # Expected behavior is deepcopy, so deepcopying array_annotations
        obj.array_annotations = deepcopy(self.array_annotations)

        obj.segment = self.segment

        return obj

    # Needed to implement this so array annotations are copied as well, ONLY WHEN copying 1:1
    def copy(self, **kwargs):
        """
        Returns a shallow copy of the object
        :return: Copy of self
        """

        obj = super().copy(**kwargs)
        obj.array_annotations = self.array_annotations
        return obj

    def as_array(self, units=None):
        """
        Return the object's data as a plain NumPy array.

        Parameters
        ----------
        units: quantities units | None, default: None
            The data return as a np.ndarray with units requested

        Returns
        -------
        data_array: np.ndarray
            The data as an array

        If `units` is specified, first rescale to those units.
        """
        if units:
            return self.rescale(units).magnitude
        else:
            return self.magnitude

    def as_quantity(self):
        """
        Return the object's data as a quantities array.
        """
        return self.view(pq.Quantity)

    def _get_arr_ann_length(self):
        """
        Return the length of the object's data as required for array annotations
        This is the last dimension of every object.
        :return Required length of array annotations for this object
        """
        # Number of items is last dimension in of data object
        # This method should be overridden in case this changes
        try:
            length = self.shape[-1]
        # Note: This is because __getitem__[int] returns a scalar Epoch/Event/SpikeTrain
        # To be removed if __getitem__[int] is changed
        except IndexError:
            length = 1
        return length

    def __deepcopy__(self, memo):
        """
        Create a deep copy of the data object.
        All attributes and annotations are also deep copied.
        References to parent objects are not kept, they are set to None.


        :param memo: (dict) Objects that have been deep copied already
        :return: (DataObject) Deep copy of the input DataObject
        """
        cls = self.__class__
        necessary_attrs = {}
        # Units need to be specified explicitly for analogsignals/irregularlysampledsignals
        for k in self._necessary_attrs + (("units",),):
            necessary_attrs[k[0]] = getattr(self, k[0], self)
        # Create object using constructor with necessary attributes
        new_obj = cls(**necessary_attrs)
        # Add all attributes
        new_obj.__dict__.update(self.__dict__)
        memo[id(self)] = new_obj
        for k, v in self.__dict__.items():
            # Single parent objects should not be deepcopied, because this is not expected behavior
            # and leads to a lot of stuff being copied (e.g. all other children of the parent as well),
            # thus creating a lot of overhead
            # But keeping the reference to the same parent is not desired either, because this would be unidirectional
            # When deepcopying top-down, e.g. a whole block, the links will be handled by the parent
            if k in self._parent_attrs:
                setattr(new_obj, k, None)
                continue
            try:
                setattr(new_obj, k, deepcopy(v, memo))
            except TypeError:
                setattr(new_obj, k, v)

        return new_obj


class ArrayDict(dict):
    """Dictionary subclass to handle array annotations

    When setting `obj.array_annotations[key]=value`, checks for consistency
    should not be bypassed.
    This class overrides __setitem__ from dict to perform these checks every time.
    The method used for these checks is given as an argument for __init__.
    """

    def __init__(self, length, check_function=_normalize_array_annotations, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_function = check_function
        self.length = length

    def __setitem__(self, key, value):
        # Directly call the defined function
        # Need to wrap key and value in a dict in order to make sure
        # that nested dicts are detected
        value = self.check_function({key: value}, self.length)[key]
        super().__setitem__(key, value)

    # Updating the dict also needs to perform checks, so rerouting this to __setitem__
    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                raise TypeError("update expected at most 1 arguments, " f"got {len(args)}")
            other = dict(args[0])
            for key in other:
                self[key] = other[key]
        for key in kwargs:
            self[key] = kwargs[key]

    def __reduce__(self):
        return super().__reduce__()



"""
baserawio
======

Classes
-------

BaseRawIO
abstract class which should be overridden to write a RawIO.

RawIO is a low level API in neo that provides fast  access to the raw data.
When possible, all IOs should/implement this level following these guidelines:
  * internal use of memmap (or hdf5)
  * fast reading of the header (do not read the complete file)
  * neo tree object is symmetric and logical: same channel/units/event
    along all block and segments.

For this level, datasets of recordings are mapped as follows:

A channel refers to a physical channel of a recording in an experiment. It is identified by a
channel_id. Recordings from a channel consist of sections of samples which are recorded
contiguously in time; in other words, a section of a channel has a specific sampling_rate,
start_time, and length (and thus also stop_time, which is the time of the sample which would
lie one sampling interval beyond the last sample present in that section).

A stream consists of a set of channels which all have the same structure of their sections of
recording and the same data type of samples. Each stream has a unique stream_id and has a name,
which does not need to be unique. A stream thus has multiple channels which all have the same
sampling rate and are on the same clock, have the same sections with t_starts and lengths, and
the same data type for their samples. The samples in a stream can thus be retrieved as an Numpy
array, a chunk of samples.

Channels within a stream can be accessed be either their channel_id, which must be unique within
a stream, or by their channel_index, which is a 0 based index to all channels within the stream.
Note that a single channel of recording may be represented within multiple streams, and such is
the case for RawIOs which may have both unfiltered and filtered or downsampled versions of the
signals from a single recording channel. In such a case, a single channel and channel_id may be
represented by a different channel_index within different streams. Lists of channel_indexes are
often convenient to pass around a selection of channels within a stream.

At the neo.io level, one AnalogSignal with multiple channels can be created for each stream. Such
an AnalogSignal may have multiple Segments, with each segment containing the sections from each
channel with the same t_start and length. Such multiple Segments for a RawIO will have the
same sampling rate. It is thus possible to retrieve the t_start and length
of the sections of the channels for a Block and Segment of a stream.

So this handles **only** one simplified but very frequent case of dataset:
    * Only one channel set  for AnalogSignal stable along Segment
    * Only one channel set  for SpikeTrain stable along Segment
    * AnalogSignal have all the same sampling_rate across all Segment
    * t_start/t_stop are the same for many object (SpikeTrain, Event) inside a Segment

Signal channels are handled by group of "stream".
One stream will result at neo.io level in one AnalogSignal with multiple channels.

A helper class `neo.io.basefromrawio.BaseFromRaw` transforms a RawIO to
neo legacy IO. In short all "neo.rawio" classes are also "neo.io"
with lazy reading capability.

With this API the IO have an attributes `header` with necessary keys.
This  `header` attribute is done in `_parse_header(...)` method.
See ExampleRawIO as example.

BaseRawIO also implements a possible persistent cache system that can be used
by some RawIOs to avoid a very long parse_header() call. The idea is that some variable
or vector can be stored somewhere (near the file, /tmp, any path) for use across multiple
constructions of a RawIO for a given set of data.

"""

from __future__ import annotations

import logging
import numpy as np
import os
import sys

from neo import logging_handler


possible_raw_modes = [
    "one-file",
    "multi-file",
    "one-dir",
]  # 'multi-dir', 'url', 'other'

error_header = "Header is not read yet, do parse_header() first"

_signal_stream_dtype = [
    ("name", "U64"),  # not necessarily unique
    ("id", "U64"),  # must be unique
]

_signal_channel_dtype = [
    ("name", "U64"),  # not necessarily unique
    ("id", "U64"),  # must be unique
    ("sampling_rate", "float64"),
    ("dtype", "U16"),
    ("units", "U64"),
    ("gain", "float64"),
    ("offset", "float64"),
    ("stream_id", "U64"),
]

# TODO for later: add t_start and length in _signal_channel_dtype
# this would simplify all t_start/t_stop stuff for each RawIO class

_common_sig_characteristics = ["sampling_rate", "dtype", "stream_id"]

_spike_channel_dtype = [
    ("name", "U64"),
    ("id", "U64"),
    # for waveform
    ("wf_units", "U64"),
    ("wf_gain", "float64"),
    ("wf_offset", "float64"),
    ("wf_left_sweep", "int64"),
    ("wf_sampling_rate", "float64"),
]

# in rawio event and epoch are handled the same way
# except, that duration is `None` for events
_event_channel_dtype = [
    ("name", "U64"),
    ("id", "U64"),
    ("type", "S5"),  # epoch or event
]


class BaseRawIO:
    """
    Generic class to handle.

    """

    name = "BaseRawIO"
    description = ""
    extensions = []

    rawmode = None  # one key from possible_raw_modes

    #   TODO Why multi-file would have a single filename is confusing here - shouldn't
    #   the name of this argument be filenames_list or filenames_base or similar?
    #
    #   When rawmode=='one-file' kargs MUST contains 'filename' the filename
    #   When rawmode=='multi-file' kargs MUST contains 'filename' one of the filenames.
    #   When rawmode=='one-dir' kargs MUST contains 'dirname' the dirname.

    def __init__(self, use_cache: bool = False, cache_path: str = "same_as_resource", **kargs):
        """
        init docstring should be filled out at the rawio level so the user knows whether to
        input filename or dirname.

        """
        # create a logger for the IO class
        fullname = self.__class__.__module__ + "." + self.__class__.__name__
        self.logger = logging.getLogger(fullname)
        # Create a logger for 'neo' and add a handler to it if it doesn't have one already.
        # (it will also not add one if the root logger has a handler)
        corename = self.__class__.__module__.split(".")[0]
        corelogger = logging.getLogger(corename)
        rootlogger = logging.getLogger()
        if not corelogger.handlers and not rootlogger.handlers:
            corelogger.addHandler(logging_handler)

        self.use_cache = use_cache
        if use_cache:
            self.setup_cache(cache_path)
        else:
            self._cache = None

        self.header = None
        self.is_header_parsed = False

    def parse_header(self):
        """
        Parses the header of the file(s) to allow for faster computations
        for all other functions

        """
        # this must create
        # self.header['nb_block']
        # self.header['nb_segment']
        # self.header['signal_streams']
        # self.header['signal_channels']
        # self.header['spike_channels']
        # self.header['event_channels']

        self._parse_header()
        self._check_stream_signal_channel_characteristics()
        self.is_header_parsed = True

    def source_name(self):
        """Return fancy name of file source"""
        return self._source_name()

    def __repr__(self):
        txt = f"{self.__class__.__name__}: {self.source_name()}\n"
        if self.header is not None:
            nb_block = self.block_count()
            txt += f"nb_block: {nb_block}\n"
            nb_seg = [self.segment_count(i) for i in range(nb_block)]
            txt += f"nb_segment:  {nb_seg}\n"

            # signal streams
            v = [
                s["name"] + f" (chans: {self.signal_channels_count(i)})"
                for i, s in enumerate(self.header["signal_streams"])
            ]
            v = pprint_vector(v)
            txt += f"signal_streams: {v}\n"

            for k in ("signal_channels", "spike_channels", "event_channels"):
                ch = self.header[k]
                v = pprint_vector(self.header[k]["name"])
                txt += f"{k}: {v}\n"

        return txt

    def _generate_minimal_annotations(self):
        """
        Helper function that generates a nested dict for annotations.

        Must be called when these are Ok after self.header is done
        and thus when these functions return the correct values:
          * block_count()
          * segment_count()
          * signal_streams_count()
          * signal_channels_count()
          * spike_channels_count()
          * event_channels_count()

        There are several sources and kinds of annotations that will
        be forwarded to the neo.io level and used to enrich neo objects:
            * annotations of objects common across segments
                * signal_streams > neo.AnalogSignal annotations
                * signal_channels > neo.AnalogSignal array_annotations split by stream
                * spike_channels > neo.SpikeTrain
                * event_channels > neo.Event and neo.Epoch
            * annotations that depend of the block_id/segment_id of the object:
              * nested in raw_annotations['blocks'][block_index]['segments'][seg_index]['signals']

        Usage after a call to this function we can do this to populate more annotations:

        raw_annotations['blocks'][block_index][ 'nickname'] = 'super block'
        raw_annotations['blocks'][block_index]
                        ['segments']['important_key'] = 'important value'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['signals']['nickname'] = 'super signals stream'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['signals']['__array_annotations__']
                        ['channels_quality'] = ['bad', 'good', 'medium', 'good']
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['spikes'][spike_chan]['nickname'] =  'super neuron'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['spikes'][spike_chan]
                        ['__array_annotations__']['spike_amplitudes'] = [-1.2, -10., ...]
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['events'][ev_chan]['nickname'] = 'super trigger'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['events'][ev_chan]
                        Z['__array_annotations__']['additional_label'] = ['A', 'B', 'A', 'C', ...]


        Theses annotations will be used at the neo.io API directly in objects.

        Standard annotation like name/id/file_origin are already generated here.
        """
        signal_streams = self.header["signal_streams"]
        signal_channels = self.header["signal_channels"]
        spike_channels = self.header["spike_channels"]
        event_channels = self.header["event_channels"]

        # use for AnalogSignal.annotations and AnalogSignal.array_annotations
        signal_stream_annotations = []
        for c in range(signal_streams.size):
            stream_id = signal_streams[c]["id"]
            channels = signal_channels[signal_channels["stream_id"] == stream_id]
            d = {}
            d["name"] = signal_streams["name"][c]
            d["stream_id"] = stream_id
            d["file_origin"] = self._source_name()
            d["__array_annotations__"] = {}
            for key in ("name", "id"):
                values = np.array([channels[key][chan] for chan in range(channels.size)])
                d["__array_annotations__"]["channel_" + key + "s"] = values
            signal_stream_annotations.append(d)

        # used for SpikeTrain.annotations and SpikeTrain.array_annotations
        spike_annotations = []
        for c in range(spike_channels.size):
            # use for Unit.annotations
            d = {}
            d["name"] = spike_channels["name"][c]
            d["id"] = spike_channels["id"][c]
            d["file_origin"] = self._source_name()
            d["__array_annotations__"] = {}
            spike_annotations.append(d)

        # used for Event/Epoch.annotations and Event/Epoch.array_annotations
        event_annotations = []
        for c in range(event_channels.size):
            # not used in neo.io at the moment could useful one day
            d = {}
            d["name"] = event_channels["name"][c]
            d["id"] = event_channels["id"][c]
            d["file_origin"] = self._source_name()
            d["__array_annotations__"] = {}
            event_annotations.append(d)

        # duplicate this signal_stream_annotations/spike_annotations/event_annotations
        # across blocks and segments and create annotations
        ann = {}
        ann["blocks"] = []
        for block_index in range(self.block_count()):
            d = {}
            d["file_origin"] = self.source_name()
            d["segments"] = []
            ann["blocks"].append(d)

            for seg_index in range(self.segment_count(block_index)):
                d = {}
                d["file_origin"] = self.source_name()
                # copy nested
                d["signals"] = signal_stream_annotations.copy()
                d["spikes"] = spike_annotations.copy()
                d["events"] = event_annotations.copy()
                ann["blocks"][block_index]["segments"].append(d)

        self.raw_annotations = ann

    def _repr_annotations(self):
        txt = "Raw annotations\n"
        for block_index in range(self.block_count()):
            bl_a = self.raw_annotations["blocks"][block_index]
            txt += f"*Block {block_index}\n"
            for k, v in bl_a.items():
                if k in ("segments",):
                    continue
                txt += f"  -{k}: {v}\n"
            for seg_index in range(self.segment_count(block_index)):
                seg_a = bl_a["segments"][seg_index]
                txt += f"  *Segment {seg_index}\n"
                for k, v in seg_a.items():
                    if k in (
                        "signals",
                        "spikes",
                        "events",
                    ):
                        continue
                    txt += f"    -{k}: {v}\n"

                # annotations by channels for spikes/events/epochs
                for child in (
                    "signals",
                    "events",
                    "spikes",
                ):
                    if child == "signals":
                        n = self.header["signal_streams"].shape[0]
                    else:
                        n = self.header[child[:-1] + "_channels"].shape[0]
                    for c in range(n):
                        neo_name = {"signals": "AnalogSignal", "spikes": "SpikeTrain", "events": "Event/Epoch"}[child]
                        txt += f"    *{neo_name} {c}\n"
                        child_a = seg_a[child][c]
                        for k, v in child_a.items():
                            if k == "__array_annotations__":
                                continue
                            txt += f"      -{k}: {v}\n"
                        for k, values in child_a["__array_annotations__"].items():
                            values = ", ".join([str(v) for v in values[:4]])
                            values = "[ " + values + " ..."
                            txt += f"      -{k}: {values}\n"

        return txt

    def print_annotations(self):
        """Print formatted raw_annotations"""
        print(self._repr_annotations())

    def block_count(self):
        """Returns the number of blocks"""
        return self.header["nb_block"]

    def segment_count(self, block_index: int):
        """
        Returns count of segments for a given block

        Parameters
        ----------
        block_index: int
            The index of the block to do the segment count for
        Returns
        -------
        count: int
            The number of segments for a given block

        """
        return self.header["nb_segment"][block_index]

    def signal_streams_count(self):
        """Return the number of signal streams.
        Same for all Blocks and Segments.
        """
        return len(self.header["signal_streams"])

    def signal_channels_count(self, stream_index: int):
        """Returns the number of signal channels for a given stream.
        This number is the same for all Blocks and Segments.

        Parameters
        ----------
        stream_index: int
            the stream index in which to count the signal channels

        Returns
        -------
        count: int
            the number of signal channels of a given stream
        """
        stream_id = self.header["signal_streams"][stream_index]["id"]
        channels = self.header["signal_channels"]
        channels = channels[channels["stream_id"] == stream_id]
        return len(channels)

    def spike_channels_count(self):
        """Return the number of unit (aka spike) channels.
        Same for all Blocks and Segments.
        """
        return len(self.header["spike_channels"])

    def event_channels_count(self):
        """Return the number of event/epoch channels.
        Same for all Blocks and Segments.
        """
        return len(self.header["event_channels"])

    def segment_t_start(self, block_index: int, seg_index: int):
        """
        Global t_start of a Segment Shared by all objects except
        for AnalogSignal.

        Parameters
        ----------
        block_index: int
            The index of the block to find the segment t_start
        seg_index: int
            The index of the segment within the block_index in which to find the t_start

        Returns
        -------
        t_start: float
            the time of global t_start of a segment within a block

        """
        return self._segment_t_start(block_index, seg_index)

    def segment_t_stop(self, block_index, seg_index):
        """
        Global t_stop of a Segment in s. Shared by all objects except
        for AnalogSignal.

        Parameters
        ----------
        block_index: int
            The index of the block to find the segment t_start
        seg_index: int
            The index of the segment within the block_index in which to find the t_start

        Returns
        -------
        t_stop: float
            the time of global t_stop of a segment within a block

        """
        return self._segment_t_stop(block_index, seg_index)

    ###
    # signal and channel zone

    def _check_stream_signal_channel_characteristics(self):
        """
        Check that all channels that belonging to the same stream_id
        have the same stream id and _common_sig_characteristics. These
        presently includes:
          * sampling_rate
          * units
          * dtype
        """
        signal_streams = self.header["signal_streams"]
        signal_channels = self.header["signal_channels"]
        if signal_streams.size > 0:
            if signal_channels.size < 1:
                raise ValueError("Signal stream exists but there are no signal channels")

        for stream_index in range(signal_streams.size):
            stream_id = signal_streams[stream_index]["id"]
            mask = signal_channels["stream_id"] == stream_id
            characteristics = signal_channels[mask][_common_sig_characteristics]
            unique_characteristics = np.unique(characteristics)
            if unique_characteristics.size != 1:
                raise ValueError(
                    f"Some channels in stream_id {stream_id} "
                    f"do not have the same {_common_sig_characteristics} {unique_characteristics}"
                )

            # also check that channel_id is unique inside a stream
            channel_ids = signal_channels[mask]["id"]
            if np.unique(channel_ids).size != channel_ids.size:
                raise ValueError(f"signal_channels do not have unique ids for stream {stream_index}")

        self._several_channel_groups = signal_streams.size > 1

    def channel_name_to_index(self, stream_index: int, channel_names: list[str]):
        """
        Inside a stream, transform channel_names to channel_indexes.
        Based on self.header['signal_channels']
        channel_indexes are zero-based offsets within the stream

        Parameters
        ----------
        stream_index: int
            The stream in which to convert channel_names to their respective channel_indexes
        channel_names: list[str]
            The channel names to convert to channel_indexes

        Returns
        -------
        channel_indexes: np.array[int]
            the channel_indexes associated with the given channel_ids

        """
        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        chan_names = list(signal_channels["name"])
        if signal_channels.size != np.unique(chan_names).size:
            raise ValueError("Channel names are not unique")
        channel_indexes = np.array([chan_names.index(name) for name in channel_names])
        return channel_indexes

    def channel_id_to_index(self, stream_index: int, channel_ids: list[str]):
        """
        Inside a stream, transform channel_ids to channel_indexes.
        Based on self.header['signal_channels']
        channel_indexes are zero-based offsets within the stream

        Parameters
        ----------
        stream_index: int
            the stream index in which to convert the channel_ids to channel_indexes
        channel_ids: list[str]
            the list of channel_ids to convert to channel_indexes

        Returns
        -------
        channel_indexes: np.array[int]
             the channel_indexes associated with the given channel_ids
        """
        # unique ids is already checked in _check_stream_signal_channel_characteristics
        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        chan_ids = list(signal_channels["id"])
        channel_indexes = np.array([chan_ids.index(chan_id) for chan_id in channel_ids])
        return channel_indexes

    def _get_channel_indexes(
        self,
        stream_index: int,
        channel_indexes: list[int] | None,
        channel_names: list[str] | None,
        channel_ids: list[str] | None,
    ):
        """
        Select channel_indexes for a stream based on channel_indexes/channel_names/channel_ids
        depending on which one is not None.

        Parameters
        ----------
        stream_index: int,
            the stream index in which to get channel_indexes
        channel_indexes: list[int] | None
            the channel_indexes desired
        channel_names: list[str] | None
            the names of the channels to be converted to channel_indexes. Give this or channel_ids for conversion
        channel_ids: list[str] | None
            the ids of the channels to be converted to channel_indexes. Give this or channel_names for conversion

        Returns
        -------
        channel_indexes: np.array[int]
            The desired channel_indexes for functions requiring channel_indexes

        """
        if channel_indexes is None and channel_names is not None:
            channel_indexes = self.channel_name_to_index(stream_index, channel_names)
        elif channel_indexes is None and channel_ids is not None:
            channel_indexes = self.channel_id_to_index(stream_index, channel_ids)
        return channel_indexes

    def _get_stream_index_from_arg(self, stream_index_arg: int | None):
        """
        Verifies the desired stream_index exists

        Parameters
        ----------
        stream_index_arg: int | None, default: None
            The stream_index to verify
            If None checks if only one stream exists and then returns 0 if it is single stream

        Returns
        -------
        stream_index: int
            The stream_index to be used for function requiring a stream_index

        """
        if stream_index_arg is None:
            if self.header["signal_streams"].size != 1:
                raise ValueError("stream_index must be given for files with multiple streams")
            stream_index = 0
        else:
            if stream_index_arg < 0 or stream_index_arg >= self.header["signal_streams"].size:
                raise ValueError(f"stream_index must be between 0 and {self.header['signal_streams'].size}")
            stream_index = stream_index_arg
        return stream_index

    def get_signal_size(self, block_index: int, seg_index: int, stream_index: int | None = None):
        """
        Retrieves the length of a single section of the channels in a stream.

        Parameters
        ----------
        block_index: int
            The desired block in which to get a signal size
        seg_index: int
            The desired segment of the block in which to get the signal size
        stream_index: int | None, default: None
            The optional stream index in which to determine signal size
            This is required for data with multiple streams

        Returns
        -------
        signal_size: int
            The number of samples for a given signal within the desired block, segment, and stream

        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        return self._get_signal_size(block_index, seg_index, stream_index)

    def get_signal_t_start(self, block_index: int, seg_index: int, stream_index: int | None = None):
        """
        Retrieves the t_start of a single section of the channels in a stream.

        Parameters
        ----------
        block_index: int
            The desired block in which to get a t_start
        seg_index: int
            The desired segment of the block in which to get the t_start
        stream_index: int | None, default: None
            The optional stream index in which to determine t_start
            This is required for data with multiple streams

        Returns
        -------
        signal_t_start: float
            The start time for a given signal within the desired block, segment, and stream

        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        return self._get_signal_t_start(block_index, seg_index, stream_index)

    def get_signal_sampling_rate(self, stream_index: int | None = None):
        """
        Retrieves the sampling rate for a stream and all channels withinin that stream.

        Parameters
        ----------
        stream_index: int | None, default: None
            The desired stream index in which to get the sampling_rate
            This is required for data with multiple streams

        Returns
        -------
        sr: float
            The sampling rate of a given stream and all channels in that stream

        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        sr = signal_channels[0]["sampling_rate"]
        return float(sr)

    def get_analogsignal_chunk(
        self,
        block_index: int = 0,
        seg_index: int = 0,
        i_start: int | None = None,
        i_stop: int | None = None,
        stream_index: int | None = None,
        channel_indexes: list[int] | None = None,
        channel_names: list[str] | None = None,
        channel_ids: list[str] | None = None,
        prefer_slice: bool = False,
    ):
        """
        Returns a chunk of raw signal as a Numpy array.

        Parameters
        ----------
        block_index: int, default: 0
            The block with the desired analog signal
        seg_index: int, default: 0
            The segment containing the desired analog signal
        i_start: int | None, default: None
            The index of the first sample (not time) of the desired analog signal
        i_stop: int | None, default: None
            The index of one past the last sample (not time) of the desired analog signal
        stream_index: int | None, default: None
            The index of the stream containing the channels to assess for the analog signal
            This is required for data with multiple streams
        channel_indexes: list[int] | np.array[int]|  slice | None, default: None
            The list of indexes of channels to retrieve
            One of channel_indexes, channel_names, or channel_ids must be given
        channel_names: list[str] | None, default: None
            The list of channel names to retrieve
            One of channel_indexes, channel_names, or channel_ids must be given
        channel_ids: list[str] | None, default: None
            The list of channel_ids to retrieve
            One of channel_indexes, channel_names, or channel_ids must be given

        Returns
        -------
        raw_chunk: np.array (n_samples, n_channels)
            The array with the raw signal samples

        Notes
        -----
        Rows are the samples and columns are the channels
        The channels are chosen either by channel_names,
        if provided, otherwise by channel_ids, if provided, otherwise by channel_indexes, if
        provided, otherwise all channels are selected.

        Examples
        --------
        # tetrode with 1 sec recording at sampling_rate = 1000. Hz
        >>> rawio_reader.parse_header()
        >>> raw_sigs = rawio_reader.get_analogsignal_chunk(block_index=2, seg_index=0, stream_index=0)
        >>> raw_sigs.shape
        (1000,4) # 1000 samples by 4 channels
        >>> raw_sigs.dtype
        'int16' # returns the dtype from the recording itself

        # If we only want one electrode
        >>> raw_sigs_one_electrode = rawio_reader.get_analogsignal_chunk(block_index=2, seg_index=0, stream_index=0, channel_indexes=[0])
        >>> raw_sigs_one_electrode.shape
        (1000,1)

        """

        signal_streams = self.header["signal_streams"]
        signal_channels = self.header["signal_channels"]
        no_signal_streams = signal_streams.size == 0
        no_channels = signal_channels.size == 0
        if no_signal_streams or no_channels:
            error_message = (
                "get_analogsignal_chunk can't be called on a file with no signal streams or channels."
                "Double check that your file contains signal streams and channels."
            )
            raise AttributeError(error_message)

        stream_index = self._get_stream_index_from_arg(stream_index)
        channel_indexes = self._get_channel_indexes(stream_index, channel_indexes, channel_names, channel_ids)

        # some check on channel_indexes
        if isinstance(channel_indexes, list):
            channel_indexes = np.asarray(channel_indexes)

        if isinstance(channel_indexes, np.ndarray):
            if channel_indexes.dtype == "bool":
                if self.signal_channels_count(stream_index) != channel_indexes.size:
                    raise ValueError(
                        "If channel_indexes is a boolean it must have be the same length as the "
                        f"number of channels {self.signal_channels_count(stream_index)}"
                    )
                (channel_indexes,) = np.nonzero(channel_indexes)

        if prefer_slice and isinstance(channel_indexes, np.ndarray):
            # Check if channel_indexes are contiguous and transform to slice argument if possible.
            # This is useful for memmap or hdf5 where providing a slice causes a lazy read,
            # rather than a list of indexes that make a copy (like numpy.take()).
            if np.all(np.diff(channel_indexes) == 1):
                channel_indexes = slice(channel_indexes[0], channel_indexes[-1] + 1)

        raw_chunk = self._get_analogsignal_chunk(block_index, seg_index, i_start, i_stop, stream_index, channel_indexes)

        return raw_chunk

    def rescale_signal_raw_to_float(
        self,
        raw_signal: np.ndarray,
        dtype: np.dtype = "float32",
        stream_index: int | None = None,
        channel_indexes: list[int] | None = None,
        channel_names: list[str] | None = None,
        channel_ids: list[str] | None = None,
    ):
        """
        Rescales a chunk of raw signals which are provided as a Numpy array. These are normally
        returned by a call to get_analogsignal_chunk.

        Parameters
        ----------
        raw_signal: np.array (n_samples, n_channels)
            The numpy array of samples with columns being samples for a single channel
        dtype: np.dype, default: "float32"
            The datatype for returning scaled samples, must be acceptable by the numpy dtype constructor
        stream_index: int | None, default: None
            The index of the stream containing the channels to assess
        channel_indexes: list[int], np.array[int], slice | None, default: None
            The list of indexes of channels to retrieve
        channel_names: list[str] | None, default: None
            The list of channel names to retrieve
        channel_ids: list[str] | None, default: None
            list of channel_ids to retrieve

        Returns
        -------
        float_signal: np.array (n_samples, n_channels)
            The rescaled signal

        Notes
        -----
        The channels are specified either by channel_names, if provided, otherwise by channel_ids,
        if provided, otherwise by channel_indexes, if provided, otherwise all channels are selected.

        These are rawio dependent because rescaling of the NumPy array requires the offset and gain
        stored within the header of the rawio


        Examples
        --------
        # Once we have a `raw_sigs` using rawio.get_analogsignal_chunk() we can convert to voltages with a desired dtype
        # If we used `stream_index=0` with `get_analogsignal_chunk` we use `stream_index=0` here
        >>> float_sigs = rawio_reader.rescale_signal_raw_to_float(raw_signal=raw_sigs, dtype='float32', stream_index=0)
        >>> float_sigs.dtype
        'float32'
        >>> float_sigs.shape
        (1000,4)
        >>> float_sigs.shape == raw_sigs.shape
        True


        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        channel_indexes = self._get_channel_indexes(stream_index, channel_indexes, channel_names, channel_ids)
        if channel_indexes is None:
            channel_indexes = slice(None)

        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        channels = self.header["signal_channels"][mask]
        if channel_indexes is None:
            channel_indexes = slice(None)
        channels = channels[channel_indexes]

        float_signal = raw_signal.astype(dtype)

        if np.any(channels["gain"] != 1.0):
            float_signal *= channels["gain"]

        if np.any(channels["offset"] != 0.0):
            float_signal += channels["offset"]

        return float_signal

    # spiketrain and unit zone
    def spike_count(self, block_index: int = 0, seg_index: int = 0, spike_channel_index: int = 0):
        """
        Returns the spike count for a given block, segment, and spike_channel_index

        Parameters
        ----------
        block_index: int, default: 0
            The block with the desired segment to assess
        seg_index: int, default: 0
            The segment containing the desired section to assess
        spike_channel_index: int, default: 0
            The spike_channel_index for assessing spike_count

        Returns
        -------
        spike_count: int
            The number of spikes in the block and segment

        """
        return self._spike_count(block_index, seg_index, spike_channel_index)

    def get_spike_timestamps(
        self,
        block_index: int = 0,
        seg_index: int = 0,
        spike_channel_index: int = 0,
        t_start: float | None = None,
        t_stop: float | None = None,
    ):
        """
        Returns the spike_timestamps in samples (see note for dtype)

        Parameters
        ----------
        block_index: int, default: 0
            The block containing the section to get the spike timestamps
        seg_index: int, default: 0
            The segment containing the section to get the spike timestamps
        spike_channel_index: int, default: 0
            The channel in which to collect spike timestamps
        t_start: float | None, default: None
            The time in seconds for the start of the section to get spike timestamps
            None indicates to start at the beginning of the segment
        t_stop: float | None, default: None
            The time in seconds for the end of the section to get spike timestamps
            None indicates to end at the end of the segment

        Returns
        -------
        timestamp: np.array
            The spike timestamps

        Notes
        -----
        The timestamp datatype is as close to the format itself. Sometimes float/int32/int64.
        Sometimes it is the index on the signal but not always.
        The conversion to second or index_on_signal is done outside this method.


        Examples
        --------
        # to look at block 1, segment 0, and channel 3 on a tetrode from 10
        # seconds to 30 seconds we would do:
        >>> timestamps = rawio_reader.get_spike_timestamps(block_index=1,
                                                           seg_index=0,
                                                           spike_channel_index=3,
                                                           t_start=10,
                                                           t_stop=30)
        """
        timestamp = self._get_spike_timestamps(block_index, seg_index, spike_channel_index, t_start, t_stop)
        return timestamp

    def rescale_spike_timestamp(self, spike_timestamps: np.ndarray, dtype: np.dtype = "float64"):
        """
        Rescale spike timestamps from samples to seconds.

        Parameters
        ----------
        spike_timestamps: np.ndarray
            The array containing the spike_timestamps to convert
        dtype: np.dtype, default: "float64"
            The dtype in which to convert the spike time in seconds. Must be accepted by the numpy.dtype constructor

        Returns
        -------
        scaled_spike_timestamps: np.array
            The spiketimes in seconds

        Examples
        --------
        # After running `get_spike_timestamps` and returning timestamps we can do the following:
        >>> scaled_spike_timestamps = rawio_reader.rescale_spike_timestamps(spike_timestamps=timestamps,
                                                                            dtype='float64')

        """
        return self._rescale_spike_timestamp(spike_timestamps, dtype)

    # spiketrain waveform zone
    def get_spike_raw_waveforms(
        self,
        block_index: int = 0,
        seg_index: int = 0,
        spike_channel_index: int = 0,
        t_start: float | None = None,
        t_stop: float | None = None,
    ):
        """
        Gets the waveforms for one channel within one segment of one block

        Parameters
        ----------
        block_index: int, default: 0
            The block containing the desired set of waveform data
        seg_index: int, default: 0
            The segment containing the desired set of waveform data
        spike_channel_index: int, default: 0
            The channel index on which to get waveform data
        t_start: float | None, default: None
            The time in seconds for the start of the section to get waveforms
            None indicates to start at the beginning of the segment
        t_stop: float | None, default: None
            The time in seconds for the end of the section to waveforms
            None indicates to end at the end of the segment

        Returns
        -------
        wf: np.ndarray (nb_spike, nb_channel, nb_sample))
            A NumPy array of spikes, channels and samples
        """
        wf = self._get_spike_raw_waveforms(block_index, seg_index, spike_channel_index, t_start, t_stop)
        return wf

    def rescale_waveforms_to_float(
        self, raw_waveforms: np.ndarray, dtype: np.dtype = "float32", spike_channel_index: int = 0
    ):
        """
        Rescale waveforms to based on the rawio's waveform gain and waveform offset

        Parameters
        ----------
        raw_waveforms: np.ndarray
            The array containing the spike_timestamps to convert
        dtype: np.dtype, default: "float64"
            The dtype in which to convert the spike time to. Must be accepted by the numpy.dtype constructor
        spike_channel_index: int, default: 0
            The channel index of the desired channel to  rescale

        Returns
        -------
        float_waveforms: np.ndarray (nb_spikes, nb_channels, nb_samples)
            The scaled waveforms to the dtype specified by dtype
        """
        wf_gain = self.header["spike_channels"]["wf_gain"][spike_channel_index]
        wf_offset = self.header["spike_channels"]["wf_offset"][spike_channel_index]

        float_waveforms = raw_waveforms.astype(dtype)

        if wf_gain != 1.0:
            float_waveforms *= wf_gain
        if wf_offset != 0.0:
            float_waveforms += wf_offset

        return float_waveforms

    # event and epoch zone
    def event_count(self, block_index: int = 0, seg_index: int = 0, event_channel_index: int = 0):
        """
        Returns the count of events for a particular block, segment, and channel_index

        Parameters
        ----------
        block_index: int, default: 0
            The block in which to count the events
        seg_index: int, default: 0
            The segment within the block given by block_index in which to count events
        event_channel_index: int, default: 0
            The index of the channel in which to count events

        Returns
        -------
        n_events: int
            The number of events in the given block, segment, and event_channel_index
        """
        return self._event_count(block_index, seg_index, event_channel_index)

    def get_event_timestamps(
        self,
        block_index: int = 0,
        seg_index: int = 0,
        event_channel_index: int = 0,
        t_start: float | None = None,
        t_stop: float | None = None,
    ):
        """
        Returns the event timestamps along with their labels and durations

        Parameters
        ----------
        block_index: int, default: 0
            The block in which to count the events
        seg_index: int, default: 0
            The segment within the block given by block_index in which to count events
        event_channel_index: int, default: 0
            The index of the channel in which to count events
        t_start: float | None, default: None
            The time in seconds for the start of the section to get waveforms
            None indicates to start at the beginning of the segment
        t_stop: float | None, default: None
            The time in seconds for the end of the section to waveforms
            None indicates to end at the end of the segment

        Returns
        -------
        timestamp: np.array
            The timestamps of events (in samples)
        durations: np.array
            The durations of each event
        labels: np.array
            The labels of the events

        Notes
        -----
        The timestamp datatype is as close to the format itself. Sometimes float/int32/int64.
        Sometimes it is the index on the signal but not always.
        The conversion to second or index_on_signal is done outside this method.

        Examples
        --------
        # A given rawio reader that generates events data. For this example we will
        # look at Block 0, Segment 1, on Channel 1, with a start time at the beginning
        # of the segment and an end time of 5 minutes (300 s)
        >>> event_timestamps, durations, labels = rawio_reader.get_event_timestamps(block_index=0,
                                                                                    seg_index=1,
                                                                                    event_channel_index=1,
                                                                                    t_start=None,
                                                                                    t_stop=300)

        """
        timestamp, durations, labels = self._get_event_timestamps(
            block_index, seg_index, event_channel_index, t_start, t_stop
        )
        return timestamp, durations, labels

    def rescale_event_timestamp(
        self, event_timestamps: np.ndarray, dtype: np.dtype = "float64", event_channel_index: int = 0
    ):
        """
        Rescale event timestamps to seconds.

        Parameters
        ----------
        event_timestamps: np.ndarray
            The array containing the event timestamps to convert
        dtype: np.dtype, default: "float64"
            The dtype in which to convert the event time in seconds. Must be accepted by the numpy.dtype constructor
        event_channel_index: int, default: 0
            The channel index for scaling the events

        Returns
        -------
        scaled_event_timestamps: np.array
            The scaled event timestamps in seconds

        Examples
        --------
        # Using the event_timestamps from the `get_event_timestamps` function we can then scale from samples into
        # seconds using this `rescale_event_timestamp`. We use the same event_channel_index as used during the
        # `get_event_timestamps`
        >>> event_timestamps_seconds = rawio_reader.rescale_event_timestamp(event_timestamps=event_timestamps,
                                                                            dtype='float64',
                                                                            event_channel_index=1)

        """
        return self._rescale_event_timestamp(event_timestamps, dtype, event_channel_index)

    def rescale_epoch_duration(
        self, raw_duration: np.ndarray, dtype: np.dtype = "float64", event_channel_index: int = 0
    ):
        """
        Rescales the epoch duration from samples to seconds

        Parameters
        ----------
        raw_duration: np.ndarray
            The array containing the epoch times in samples
        dtype: np.dtype, default: "float64"
            The dtype in which to convert the spike time in seconds. Must be accepted by the numpy.dtype constructor
        event_channel_index: int, default: 0
            The channel on which to index for scaling epochs

        Returns
        -------
        scaled_epoch_durations: np.array
            The scaled epoch durations in seconds

        Examples
        --------
        # In this example we use the durations obtained from running `get_event_timestamps`
        >>> duration_seconds = rawio_reader.rescale_epoch_duration(raw_durations=durations,
                                                                   dtype='float64',
                                                                   event_channel_index=0)
        """
        return self._rescale_epoch_duration(raw_duration, dtype, event_channel_index)

    def setup_cache(self, cache_path: "home" | "same_as_resource", **init_kargs):
        try:
            import joblib
        except ImportError:
            raise ImportError("Using the RawIO cache needs joblib to be installed")

        if self.rawmode in ("one-file", "multi-file"):
            resource_name = self.filename
        elif self.rawmode == "one-dir":
            resource_name = self.dirname
        else:
            raise (NotImplementedError)

        if cache_path == "home":
            if sys.platform.startswith("win"):
                dirname = os.path.join(os.environ["APPDATA"], "neo_rawio_cache")
            elif sys.platform.startswith("darwin"):
                dirname = "~/Library/Application Support/neo_rawio_cache"
            else:
                dirname = os.path.expanduser("~/.config/neo_rawio_cache")
            dirname = os.path.join(dirname, self.__class__.__name__)

            if not os.path.exists(dirname):
                os.makedirs(dirname)
        elif cache_path == "same_as_resource":
            dirname = os.path.dirname(resource_name)
        else:
            if not os.path.exists(cache_path):
                raise ValueError("cache_path does not exist use 'home' or 'same_as_resource' to make this auto")

        # the hash of the resource (dir of file) is done with filename+datetime
        # TODO make something more sophisticated when rawmode='one-dir' that use all
        #  filename and datetime
        d = dict(ressource_name=resource_name, mtime=os.path.getmtime(resource_name))
        hash = joblib.hash(d, hash_name="md5")

        # name is constructed from the resource_name and the hash
        name = f"{os.path.basename(resource_name)}_{hash}"
        self.cache_filename = os.path.join(dirname, name)

        if os.path.exists(self.cache_filename):
            self.logger.warning(f"Use existing cache file {self.cache_filename}")
            self._cache = joblib.load(self.cache_filename)
        else:
            self.logger.warning(f"Create cache file {self.cache_filename}")
            self._cache = {}
            self.dump_cache()

    def add_in_cache(self, **kargs):
        if not self.use_cache:
            raise ValueError("Can not use add_in_cache if not using cache")
        self._cache.update(kargs)
        self.dump_cache()

    def dump_cache(self):
        if not self.use_cache:
            raise ValueError("Can not use dump_cache if not using cache")
        joblib.dump(self._cache, self.cache_filename)

    ##################

    # Functions to be implemented in IO below here

    def _parse_header(self):
        raise (NotImplementedError)
        # must call
        # self._generate_empty_annotations()

    def _source_name(self):
        raise (NotImplementedError)

    def _segment_t_start(self, block_index: int, seg_index: int):
        raise (NotImplementedError)

    def _segment_t_stop(self, block_index: int, seg_index: int):
        raise (NotImplementedError)

    ###
    # signal and channel zone
    def _get_signal_size(self, block_index: int, seg_index: int, stream_index: int):
        """
        Return the size of a set of AnalogSignals indexed by channel_indexes.

        All channels indexed must have the same size and t_start.
        """
        raise (NotImplementedError)

    def _get_signal_t_start(self, block_index: int, seg_index: int, stream_index: int):
        """
        Return the t_start of a set of AnalogSignals indexed by channel_indexes.

        All channels indexed must have the same size and t_start.
        """
        raise (NotImplementedError)

    def _get_analogsignal_chunk(
        self,
        block_index: int,
        seg_index: int,
        i_start: int | None,
        i_stop: int | None,
        stream_index: int,
        channel_indexes: list[int] | None,
    ):
        """
        Return the samples from a set of AnalogSignals indexed
        by stream_index and channel_indexes (local index inner stream).

        RETURNS
        -------
            array of samples, with each requested channel in a column
        """

        raise (NotImplementedError)

    ###
    # spiketrain and unit zone
    def _spike_count(self, block_index: int, seg_index: int, spike_channel_index: int):
        raise (NotImplementedError)

    def _get_spike_timestamps(
        self, block_index: int, seg_index: int, spike_channel_index: int, t_start: float | None, t_stop: float | None
    ):
        raise (NotImplementedError)

    def _rescale_spike_timestamp(self, spike_timestamps: np.ndarray, dtype: np.dtype):
        raise (NotImplementedError)

    ###
    # spike waveforms zone
    def _get_spike_raw_waveforms(
        self, block_index: int, seg_index: int, spike_channel_index: int, t_start: float | None, t_stop: float | None
    ):
        raise (NotImplementedError)

    ###
    # event and epoch zone
    def _event_count(self, block_index: int, seg_index: int, event_channel_index: int):
        raise (NotImplementedError)

    def _get_event_timestamps(
        self, block_index: int, seg_index: int, event_channel_index: int, t_start: float | None, t_stop: float | None
    ):
        raise (NotImplementedError)

    def _rescale_event_timestamp(self, event_timestamps: np.ndarray, dtype: np.dtype):
        raise (NotImplementedError)

    def _rescale_epoch_duration(self, raw_duration: np.ndarray, dtype: np.dtype):
        raise (NotImplementedError)


def pprint_vector(vector, lim: int = 8):
    vector = np.asarray(vector)
    if vector.ndim != 1:
        raise ValueError(f"`vector` must have a dimension of 1 and not {vector.ndim}")
    if len(vector) > lim:
        part1 = ", ".join(e for e in vector[: lim // 2])
        part2 = " , ".join(e for e in vector[-lim // 2 :])
        txt = f"[{part1} ... {part2}]"
    else:
        part1 = ", ".join(e for e in vector)
        txt = f"[{part1}]"
    return txt
